//=============================================================================
//
//   File : kvi_ircconnection.cpp
//   Created on Mon 03 May 2004 01:45:42 by Szymon Stefanek
//
//   This file is part of the KVIrc IRC client distribution
//   Copyright (C) 2004 Szymon Stefanek <pragma at kvirc dot net>
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#include "kvi_ircconnection.h"

#include "kvi_dns.h"
#include "kvi_locale.h"
#include "kvi_ircserverdb.h"
#include "kvi_proxydb.h"
#include "kvi_error.h"
#include "kvi_out.h"
#include "kvi_options.h"
#include "kvi_irclink.h"
#include "kvi_console.h"
#include "kvi_netutils.h"
#include "kvi_internalcmd.h"
#include "kvi_frame.h"
#include "kvi_mexlinkfilter.h"
#include "kvi_garbage.h"

#include <qtimer.h>

extern KVIRC_API KviIrcServerDataBase           * g_pIrcServerDataBase;
extern KVIRC_API KviProxyDataBase               * g_pProxyDataBase;
extern KVIRC_API KviGarbageCollector            * g_pGarbageCollector;


KviIrcConnection::KviIrcConnection(KviConsole * pConsole,
				const QString &szNetworkName,
				KviIrcServer * pServer,
				KviProxy * pProxy)
: QObject()
{
	m_pConsole = pConsole;
	m_szNetworkName = szNetworkName;
	m_pServer = new KviIrcServer(*m_pServer);
	m_pProxy = pProxy ? new KviProxy(*m_pProxy) : 0;
	m_pLink = 0;
	m_pStartTimer = 0;
	m_pProxyDns = 0;
	m_pServerDns = 0;
}

KviIrcConnection::~KviIrcConnection()
{
	if(m_pProxyDns)delete m_pProxyDns;
	if(m_pServerDns)delete m_pServerDns;
	if(m_pLink)destroyLink();
	delete m_pServer;
	if(m_pProxy)delete m_pProxy;
	if(m_pStartTimer)delete m_pStartTimer;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// KviIrcLink management
//

void KviIrcConnection::linkDestroyed()
{
	// ops.. the link has been destroyed without permission :D
	m_pLink = 0;
	m_pConsole->output(KVI_OUT_SYSTEMWARNING,__tr2qs("Ops... for some reason the link object has been destroyed"));
	createLink(QString::null); // we need a new one
}

void KviIrcConnection::destroyLink()
{
	if(!m_pLink)return;
	// disconnect , in particular kill the destroyed() signal
	disconnect(m_pLink,0,this,0);
	// the module extension server links must be destroyed in the module that provided it
	if(m_pLink->inherits("KviMexLinkFilter"))((KviMexLinkFilter *)m_pLink)->die();
	else delete m_pLink; // we can simply delete the builtin objects
	m_pLink = 0;
}

void KviIrcConnection::createLink(const QString &szLinkFilterName)
{
	if(m_pLink)destroyLink();

	if(szLinkFilterName.isEmpty())goto alloc_builtin_link;
	if(KviQString::equalCI(szLinkFilterName,"irc"))goto alloc_builtin_link;

	m_pLink = (KviIrcLink *)g_pModuleExtensionManager->allocateExtension("linkfilter",
								szLinkFilterName.latin1(),m_pConsole,0,0,szLinkFilterName.latin1());

	if(m_pLink)
	{
		connect(m_pLink,SIGNAL(destroyed()),this,SLOT(linkDestroyed()));
		m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
				__tr2wc("Using filtered IRC protocol: Link filter is \"%Q\""),&szLinkFilterName);
		return;
	}

	m_pConsole->output(KVI_OUT_SYSTEMWARNING,
			__tr2wc("Failed to set up the link filter \"%Q\", will try with plain IRC"),&szLinkFilterName);

alloc_builtin_link:
	m_pLink = new KviIrcLink(m_pConsole);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Connection operations
//

void KviIrcConnection::start()
{
	if(m_pStartTimer) // this should never happen I guess
	{
		delete m_pStartTimer;
		m_pStartTimer = 0;
	}
	m_pStartTimer = new QTimer(this);
	connect(m_pStartTimer,SIGNAL(timeout()),this,SLOT(startConnection()));
	m_pStartTimer->start(0);
}

void KviIrcConnection::startConnection()
{
	if(m_pStartTimer)
	{
		delete m_pStartTimer;
		m_pStartTimer = 0;
	}

	m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
		__tr2wc("Attempting %w to %s (%s) on port %u"),
		m_pServer->useSSL() ? __tr2wc("secure connection") : __tr2wc("connection"),
		m_pServer->m_szHostname.latin1(),
		m_szNetworkName.latin1(),
		m_pServer->m_uPort);

	if(m_pProxy)
	{
		m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
			__tr2wc("Attempting to 'bounce' on proxy %s on port %u (protocol %s)"),
			m_pProxy->m_szHostname.ptr(),
			m_pProxy->m_uPort,
			m_pProxy->protocolName());

		lookupProxyHostname();
	} else {
		lookupServerHostname();
	}
}

void KviIrcConnection::lookupProxyHostname()
{
	bool bValidIp;

#ifdef COMPILE_IPV6_SUPPORT
	if(m_pProxy->isIpV6())
	{
		bValidIp = kvi_isValidStringIp_V6(m_pProxy->m_szIp.ptr());
	} else {
#endif
		bValidIp = kvi_isValidStringIp(m_pProxy->m_szIp.ptr());
#ifdef COMPILE_IPV6_SUPPORT
	}
#endif

	if(bValidIp)
	{
		m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
			__tr2wc("Using cached proxy IP address (%s)"),
			m_pProxy->m_szIp.ptr());
		lookupServerHostname();
	} else {
		if(m_pProxyDns)
		{
			debug("Something weird is happening, m_pProxyDns is non-zero in lookupProxyHostname()");
			delete m_pProxyDns;
		}
		m_pProxyDns = new KviDns();
		connect(m_pProxyDns,SIGNAL(lookupDone(KviDns *)),this,SLOT(proxyLookupTerminated(KviDns *)));

		if(!m_pProxyDns->lookup(m_pProxy->m_szHostname.ptr(),
			m_pProxy->isIpV6() ? KviDns::IpV6 : KviDns::IpV4))
		{
			m_pConsole->outputNoFmt(KVI_OUT_SYSTEMWARNING,
					__tr2wc("Unable to look up the IRC proxy hostname: Can't start the DNS slave"));
			m_pConsole->outputNoFmt(KVI_OUT_SYSTEMWARNING,
					__tr2wc("Resuming direct server connection"));
			// FIXME: #warning "Option for resuming direct connection or not ?"
			delete m_pProxyDns;
			m_pProxyDns = 0;
			delete m_pProxy;
			m_pProxy = 0;
			lookupServerHostname();
		} else {
			m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
				__tr2wc("Looking up the proxy hostname (%s)..."),
				m_pProxy->m_szHostname.ptr());
		}
	}
}

void KviIrcConnection::lookupServerHostname()
{
	bool bValidIp;

#ifdef COMPILE_IPV6_SUPPORT
	if(m_pServer->isIpV6())
	{
		bValidIp = kvi_isValidStringIp_V6(m_pServer->m_szIp.latin1());
	} else {
#endif
		bValidIp = kvi_isValidStringIp(m_pServer->m_szIp.latin1());
#ifdef COMPILE_IPV6_SUPPORT
	}
#endif

	if(bValidIp && m_pServer->cacheIp())
	{
		m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
			__tr2wc("Using cached server IP address (%s)"),
			m_pServer->m_szIp.latin1());
		haveServerIp();
	} else {
		if(m_pServerDns)
		{
			debug("Something weird is happening, m_pServerDns is non-zero in lookupServerHostname()");
			delete m_pServerDns;
		}
		m_pServerDns = new KviDns();
		connect(m_pServerDns,SIGNAL(lookupDone(KviDns *)),this,
			SLOT(serverLookupTerminated(KviDns *)));
		if(!m_pServerDns->lookup(m_pServer->m_szHostname.latin1(),
			m_pServer->isIpV6() ? KviDns::IpV6 : KviDns::IpV4))
		{
			m_pConsole->outputNoFmt(KVI_OUT_SYSTEMERROR,
				__tr2wc("Unable to look up the server hostname: Can't start the DNS slave"));
			emit connectionFailed();
		} else {
			m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
				__tr2wc("Looking up the server hostname (%s)..."),
				m_pServer->m_szHostname.latin1());
		}
	}
}


void KviIrcConnection::proxyLookupTerminated(KviDns *)
{
	if(m_pProxyDns->state() != KviDns::Success)
	{
		m_pConsole->output(KVI_OUT_SYSTEMERROR,
			__tr2wc("Can't find the proxy IP address: %s"),
			kvi_getErrorString(m_pProxyDns->error()));
		// FIXME: #warning "Option to resume the direct connection if proxy failed ?"
		m_pConsole->output(KVI_OUT_SYSTEMERROR,
			__tr2wc("Resuming direct server connection"));
		delete m_pProxy;
		m_pProxy = 0;
	} else {
		m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
			__tr2wc("Proxy hostname resolved to %s"),m_pProxyDns->firstIpAddress());
		m_pProxy->m_szIp = m_pProxyDns->firstIpAddress();
		g_pProxyDataBase->updateProxyIp(m_pProxy->m_szIp.ptr(),m_pProxyDns->firstIpAddress());

		if(m_pProxyDns->hostnameCount() > 1)
		{
			KviStr szFirstHostname = m_pProxyDns->firstHostname();

			for(KviStr * addr = m_pProxyDns->hostnameList()->next();addr;addr = m_pProxyDns->hostnameList()->next())
			{
				m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
					__tr2wc("Proxy %s has a nickname: %s"),szFirstHostname.ptr(),addr->ptr());
			}
		}
	}

	delete m_pProxyDns;
	m_pProxyDns = 0;
	// FIXME: #warning "With http proxies we might even avoid lookups for server names!"
	lookupServerHostname();
}


void KviIrcConnection::serverLookupTerminated(KviDns *)
{
	if(m_pServerDns->state() != KviDns::Success)
	{
		m_pConsole->output(KVI_OUT_SYSTEMERROR,
			__tr2wc("Can't find the server IP address: %s"),
			kvi_getErrorString(m_pServerDns->error()));
		m_pConsole->output(KVI_OUT_SYSTEMERROR,
			__tr2wc("Connection attempt failed [%s]"),
			m_pServer->m_szHostname.latin1());

#ifdef COMPILE_IPV6_SUPPORT
		if(!(m_pServer->isIpV6()))
		{
			m_pConsole->output(KVI_OUT_SYSTEMERROR,
				__tr2wc("If this server is an IPv6 one, try /server -i %s"),
				m_pServer->m_szHostname.latin1());
		}
#endif
		emit connectionFailed();
		return;
	}
	
	m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
		__tr2wc("Server hostname resolved to %s"),
		m_pServerDns->firstIpAddress());
	g_pIrcServerDataBase->updateServerIp(m_pServer->m_szHostname,
		m_pServer->m_uPort,m_pServerDns->firstIpAddress());

	if(!kvi_strEqualCI(m_pServer->m_szHostname.latin1(),m_pServerDns->firstHostname()))
	{
		m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
			__tr2wc("Real hostname for %s is %s"),
			m_pServer->m_szHostname.latin1(),
			m_pServerDns->firstHostname());
		m_pServer->m_szHostname = m_pServerDns->firstHostname();
	}

	m_pServer->m_szIp = m_pServerDns->firstIpAddress();

	if(m_pServerDns->hostnameCount() > 1)
	{
		KviStr szFirstHostname = m_pServerDns->firstHostname();

		for(KviStr * addr = m_pServerDns->hostnameList()->next();addr;addr = m_pServerDns->hostnameList()->next())
		{
			m_pConsole->output(KVI_OUT_SYSTEMMESSAGE,
				__tr2wc("Server %s has a nickname: %s"),
				szFirstHostname.ptr(),addr->ptr());
		}
	}

	delete m_pServerDns;
	m_pServerDns = 0;
	haveServerIp();
}

bool KviIrcConnection::validateLocalAddress(const QString &szAddress,QString &szBuffer)
{
	// szAddress may be an ip address or an interface name
#ifdef COMPILE_IPV6_SUPPORT
	if(m_pServer->isIpV6())
	{
		if(kvi_isValidStringIp_V6(szAddress.latin1()))
		{
			szBuffer = szAddress;
			return true;
		}
	} else {
#endif
		if(KviNetUtils::isValidStringIp(szAddress))
		{
			szBuffer = szAddress;
			return true;
		}
#ifdef COMPILE_IPV6_SUPPORT
	}
#endif

	// is it an interface name ?
	return KviNetUtils::getInterfaceAddress(szAddress,szBuffer);
}


void KviIrcConnection::haveServerIp()
{
	if(KVI_OPTION_BOOL(KviOption_boolUseIdentService))
		m_pConsole->frame()->executeInternalCommand(KVI_INTERNALCOMMAND_IDENT_START);

	QString bindAddress;

	if(!m_szBindAddress.isEmpty())
	{
		if(!validateLocalAddress(m_szBindAddress,bindAddress))
		{
			if((m_szBindAddress.find('.') != -1) ||
				(m_szBindAddress.find(':') != -1))
			{
				m_pConsole->output(KVI_OUT_SYSTEMWARNING,
					__tr2wc("The specified bind address (%Q) is not valid"),
					&m_szBindAddress);
			} else {
				m_pConsole->output(KVI_OUT_SYSTEMWARNING,
					__tr2wc("The specified bind address (%Q) is not valid (the interface it refers to might be down)"),
					&m_szBindAddress);
			}
			m_szBindAddress = "";
		}
	} else {
		// the options specify a bind address ?
#ifdef COMPILE_IPV6_SUPPORT
		if(m_pServer->isIpV6())
		{
			if(KVI_OPTION_BOOL(KviOption_boolBindIrcIpV6ConnectionsToSpecifiedAddress))
			{
				if(KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).hasData())
				{
					if(!validateLocalAddress(QString(KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).ptr()),bindAddress))
					{
						// if it is not an interface name , kill it for now and let the user correct the address
						if(KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).findFirstIdx(':') != -1)
						{
							m_pConsole->output(KVI_OUT_SYSTEMWARNING,
								__tr2wc("The system-wide IPv6 bind address (%s) is not valid"),
								KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).ptr());
							KVI_OPTION_BOOL(KviOption_boolBindIrcIpV6ConnectionsToSpecifiedAddress) = false;
						} else {
							// this is an interface address: might be down
							m_pConsole->output(KVI_OUT_SYSTEMWARNING,
								__tr2wc("The system-wide IPv6 bind address (%s) is not valid (the interface it refers to might be down)"),
								KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).ptr());
						}
					}
				} else {
					// empty address....kill it
					KVI_OPTION_BOOL(KviOption_boolBindIrcIpV6ConnectionsToSpecifiedAddress) = false;
				}
			}
		} else {
#endif
			if(KVI_OPTION_BOOL(KviOption_boolBindIrcIpV4ConnectionsToSpecifiedAddress))
			{
				if(KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).hasData())
				{
					if(!validateLocalAddress(QString(KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).ptr()),bindAddress))
					{
						// if it is not an interface name , kill it for now and let the user correct the address
						if(KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).findFirstIdx(':') != -1)
						{
							m_pConsole->output(KVI_OUT_SYSTEMWARNING,
								__tr2wc("The system-wide IPv4 bind address (%s) is not valid"),
								KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).ptr());
							KVI_OPTION_BOOL(KviOption_boolBindIrcIpV4ConnectionsToSpecifiedAddress) = false;
						} else {
							// this is an interface address: might be down
							m_pConsole->output(KVI_OUT_SYSTEMWARNING,
								__tr2wc("The system-wide IPv4 bind address (%s) is not valid (the interface it refers to might be down)"),
								KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).ptr());
						}
					}
				} else {
					// empty address....kill it
					KVI_OPTION_BOOL(KviOption_boolBindIrcIpV4ConnectionsToSpecifiedAddress) = false;
				}
			}
#ifdef COMPILE_IPV6_SUPPORT
		}
#endif
	}

	createLink(m_pServer->linkFilter());

	int iErr = m_pLink->startConnection(m_pServer,m_pProxy,
					bindAddress.isEmpty() ? 0 : bindAddress.latin1());

	if(iErr != KviError_success)
	{
		m_pConsole->output(KVI_OUT_SYSTEMERROR,
			__tr2wc("Connection attempt failed (%s): %s"),
			m_pServer->m_szHostname.latin1(),
			kvi_getErrorString(iErr));

		emit connectionFailed();
	}
}

