//=============================================================================
//
//   File : kvi_console.cpp
//   Creation date : Sun Jun 25 2000 15:01:34 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2004 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#include "kvi_debug.h"
#include "kvi_app.h"
#include "kvi_console.h"
#include "kvi_frame.h"
#include "kvi_iconmanager.h"
#include "kvi_options.h"
#include "kvi_locale.h"
#include "kvi_ircview.h"
#include "kvi_mirccntrl.h"
#include "kvi_out.h"
#include "kvi_input.h"
#include "kvi_error.h"
#include "kvi_ircserverdb.h"
#include "kvi_proxydb.h"
#include "kvi_netutils.h"
#include "kvi_ircserver.h"
#include "kvi_dns.h"
#include "kvi_defaults.h"
#include "kvi_ircuserdb.h"
#include "kvi_channel.h"
#include "kvi_query.h"
#include "kvi_event.h"
#include "kvi_parameterlist.h"
#include "kvi_uparser.h"
#include "kvi_regusersdb.h"
#include "kvi_userlistview.h"
#include "kvi_notifylist.h"
#include "kvi_config.h"
#include "kvi_irctoolbar.h"
#include "kvi_internalcmd.h"
#include "kvi_sparser.h"
#include "kvi_themedlabel.h"
#include "kvi_garbage.h"
#include "kvi_modulemanager.h"
#include "kvi_fileutils.h"
#include "kvi_irctoolbar.h"
#include "kvi_time.h"
#include "kvi_mexlinkfilter.h"
#include "kvi_avatarcache.h"

#ifdef COMPILE_SSL_SUPPORT
	#include "kvi_sslmaster.h"
#endif

#include <qsplitter.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qtimer.h>
#include <qhbox.h>
#include <qpopupmenu.h>
#include <qmessagebox.h>
#include <qstringlist.h>

extern KVIRC_API KviIrcServerDataBase           * g_pIrcServerDataBase;
extern KVIRC_API KviProxyDataBase               * g_pProxyDataBase;
extern KVIRC_API KviRegisteredUserDataBase      * g_pRegisteredUserDataBase;
extern KVIRC_API KviGarbageCollector            * g_pGarbageCollector;



KviConnectionInfo::KviConnectionInfo(KviIrcNetwork * net,KviIrcServer * srv,KviProxy * prx)
{

	szNetworkName            = net->name();
	pIrcServer               = new KviIrcServer(*srv);
	pProxy                   = prx ? new KviProxy(*prx) : 0;

	pNotifyListManager       = 0;
	bSentQuit                = false;
	pDns                     = 0;
	pUserDataBase            = new KviIrcUserDataBase();
	pChannelList             = new KviPtrList<KviChannel>;
	pChannelList->setAutoDelete(false);
	pQueryList               = new KviPtrList<KviQuery>;
	pQueryList->setAutoDelete(false);
	pWhoisInfoList           = new KviPtrList<KviWhoisInfo>;
	pWhoisInfoList->setAutoDelete(true);
	lastNetsplitTime         = 0;
	lastCtcpTime             = 0;
	uCtcpCount               = 0;
	szServerSupportedChanTypes = "#&!+";
	szServerSupportedModePrefixes = "@+";
	bServerSupportsModeIe    = true;
	bServerSupportsWatchList = false;
	pNotifyListTimer         = 0;
	bAway                    = false;
	tConnectionStart         = 0;
	pLocalhostDns            = 0;
}

KviConnectionInfo::~KviConnectionInfo()
{
	if(pNotifyListTimer)
	{
		delete pNotifyListTimer;
		pNotifyListTimer = 0;
	}
	if(pIrcServer)delete pIrcServer;
	if(pProxy)delete pProxy;
	if(pDns)
	{
		QObject::disconnect(pDns,SIGNAL(lookupDone(KviDns *)),0,0);
		if(pDns->isRunning())
		{
			// the DNS thread is running in the background... ugh
			// we need to wait it to finish before deleting it if we don't want to be blocked
			g_pGarbageCollector->collect(pDns);
		} else {
			// this is never blocking.. just delete the object
			delete pDns;
		}
	}
	if(pLocalhostDns)
	{
		QObject::disconnect(pLocalhostDns,SIGNAL(lookupDone(KviDns *)),0,0);
		if(pLocalhostDns->isRunning())
		{
			g_pGarbageCollector->collect(pLocalhostDns);
		} else {
			delete pLocalhostDns;
		}
	}
	if(pNotifyListManager)delete pNotifyListManager; // destroy this before the userDb
	delete pUserDataBase;
	delete pChannelList;
	delete pQueryList;
	delete pWhoisInfoList;
}

KviChannel * KviConnectionInfo::findChannel(const char * name)
{
	for(KviChannel * c = pChannelList->first();c;c = pChannelList->next())
	{
		if(kvi_strEqualCI(name,c->name()))return c;
	}
	return 0;
}

KviQuery * KviConnectionInfo::findQuery(const char * nick)
{
	for(KviQuery * q = pQueryList->first();q;q = pQueryList->next())
	{
		if(q->haveTarget(nick))return q;
	}
	return 0;
}













KviConsole::KviConsole(KviFrame * lpFrm,int iFlags)
#ifdef COMPILE_ON_WINDOWS
: KviWindow(KVI_WINDOW_TYPE_CONSOLE,lpFrm,"CONSOLE",0)
#else
: KviWindow(KVI_WINDOW_TYPE_CONSOLE,lpFrm,"CONSOLE",this)
#endif
{
	m_pConsole = this;
	m_pIrcContext = new KviIrcContext(this);

	m_iFlags = iFlags;
	if(m_pIrcContext->id() == 1)
	{
		m_iFlags |= KVI_CONSOLE_FLAG_FIRSTINAPP;
	}


	m_pMonitorList = 0;
	m_uConnectAttemptCount = 0;
	m_pConnectionInfo = 0;
	m_pLastIrcServer = 0;

	m_pLink = 0;
	createLink(KviStr::emptyString());

	m_pButtonBox = new QHBox(this,"button_box");
	KviThemedLabel * l = new KviThemedLabel(m_pButtonBox,"dummy_label");
	m_pButtonBox->setStretchFactor(l,1);
	QToolTip::add(l,__tr2qs("What's this?"));

	m_pNotifyViewButton = createToolButton(m_pButtonBox,"list_view_button",KVI_SMALLICON_HIDELISTVIEW,KVI_SMALLICON_SHOWLISTVIEW,__tr2qs("Toggle Notify List"),true);
	connect(m_pNotifyViewButton,SIGNAL(clicked()),this,SLOT(toggleNotifyView()));

	m_pSplitter = new QSplitter(QSplitter::Horizontal,this,"splitter");
	m_pIrcView = new KviIrcView(m_pSplitter,lpFrm,this);
	connect(m_pIrcView,SIGNAL(rightClicked()),this,SLOT(textViewRightClicked()));

	// FIXME: #warning "If notify list is disabled avoid to show this"
	// FIXME: #warning "Button to show/hide the notifyListView (NOT DELETE RE_CREATE!)"
	// The userlist on the right
	m_pNotifyListView = new KviUserListView(m_pSplitter,0,this,19,KviWStr (__tr2wc("Notify List")).qstring(),"notify_list_view");

	m_pAsyncServerCommandData = 0;
	m_pInput   = new KviInput(this,m_pNotifyListView);

	m_state = NotConnected;

	m_pLagMeter = new KviLagMeter(this);

	if(KVI_OPTION_BOOL(KviOption_boolAutoLogConsole))m_pIrcView->startLogging(0);
}



KviConsole::~KviConsole()
{
	// FIXME: #warning "WARNING : THIS SHOULD BECOME A COMMAND /QUIT $option() so the idents are parsed!"
	while(m_pMonitorList)
	{
		KviIrcLinkMonitor * m = m_pMonitorList->first();
		if(m)
		{
			m->die();
		} else {
			delete m_pMonitorList;
			m_pMonitorList = 0;
		}
	}

	if(m_pLagMeter->isEnabled())m_pLagMeter->disable();
	delete m_pLagMeter;
	m_pLagMeter = 0;

	// Force connection close
	if(m_state == Connected)terminateConnectionRequest(true);

	__range_valid(m_state == NotConnected);
	__range_valid(m_pConnectionInfo == 0);

	TRIGGER_EVENT(KviEvent_OnIrcContextDestroyed,this);

	if(m_pFrm->consoleCount() <= 1)
	{
		TRIGGER_EVENT(KviEvent_OnFrameWindowDestroyed,this);
		if(g_pApp->frameCount() < 1) TRIGGER_EVENT(KviEvent_OnKVIrcShutdown,this);
	}

	if(m_pLastIrcServer)delete m_pLastIrcServer;

	destroyLink();

	destroyAsyncServerCommandData();

	delete m_pIrcContext;
}


void KviConsole::registerLinkMonitor(KviIrcLinkMonitor * m)
{
	if(!m_pMonitorList)
	{
		m_pMonitorList = new KviPtrList<KviIrcLinkMonitor>;
		m_pMonitorList->setAutoDelete(false);
	}
	m_pMonitorList->append(m);
}

void KviConsole::unregisterLinkMonitor(KviIrcLinkMonitor *m)
{
	if(!m_pMonitorList)return;
	m_pMonitorList->removeRef(m);
	if(m_pMonitorList->isEmpty())
	{
		delete m_pMonitorList;
		m_pMonitorList = 0;
	}
}

void KviConsole::destroyConnection()
{
	if(m_pConnectionInfo)
	{
		closeAllChannels();
		closeAllQueries();
		if(m_pLagMeter)
		{
			if(m_pLagMeter->isEnabled())m_pLagMeter->disable();
		}
		m_pNotifyListView->partAll();
		m_pNotifyListView->setUserDataBase(0); // this is rather for crash tests
		if(ircContext()->linksWindow())ircContext()->linksWindow()->control(EXTERNAL_SERVER_DATA_PARSER_CONTROL_RESET);
		if(ircContext()->listWindow())ircContext()->listWindow()->control(EXTERNAL_SERVER_DATA_PARSER_CONTROL_RESET);
		delete m_pConnectionInfo;
		m_pConnectionInfo = 0;
	}
}

void KviConsole::triggerCreationEvents()
{
	if(m_iFlags & KVI_CONSOLE_FLAG_FIRSTINAPP) // this is the first context in the application
	{
		TRIGGER_EVENT(KviEvent_OnKVIrcStartup,this);

		if(KVI_OPTION_BOOL(KviOption_boolShowTipAtStartup))
			m_pFrm->executeInternalCommand(KVI_INTERNALCOMMAND_TIP_OPEN);

		if(KVI_OPTION_BOOL(KviOption_boolShowServersConnectDialogOnStart))
			m_pFrm->executeInternalCommand(KVI_INTERNALCOMMAND_SERVERSJOIN_OPEN);
	}

	if(m_iFlags & KVI_CONSOLE_FLAG_FIRSTINFRAME)
	{
		TRIGGER_EVENT(KviEvent_OnFrameWindowCreated,this);
	}

	TRIGGER_EVENT(KviEvent_OnIrcContextCreated,this);
}


void KviConsole::fillContextPopup(QPopupMenu * p)
{
	p->insertSeparator();
	int id = p->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CHANNEL)),__tr2qs("Part All Channels"),this,SLOT(partAllChannels()));
	if(!channelCount())p->setItemEnabled(id,false);
	id = p->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_QUERY)),__tr2qs("Close All Queries"),this,SLOT(closeAllQueries()));
	if(!queryCount())p->setItemEnabled(id,false);
}

void KviConsole::completeChannel(const KviStr &word,KviPtrList<KviStr> * matches)
{
	for(QStringList::Iterator it = KVI_OPTION_STRINGLIST(KviOption_stringlistRecentChannels).begin(); it != KVI_OPTION_STRINGLIST(KviOption_stringlistRecentChannels).end(); ++it)
	{
		if(kvi_strEqualCIN( (*it).ascii(), word, word.len()))matches->append(new KviStr((*it).ascii()));
	}
}

void KviConsole::completeServer(const KviStr &word, KviPtrList<KviStr> * matches)
{
	for(QStringList::Iterator it = KVI_OPTION_STRINGLIST(KviOption_stringlistRecentServers).begin(); it != KVI_OPTION_STRINGLIST(KviOption_stringlistRecentServers).end(); ++it)
	{
		KviStr srv((*it));

		srv.cutToFirst('/');
		while(srv.firstCharIs('/'))srv.cutLeft(1);
		srv.cutFromLast(':');
		//We should have a full server name here, without the irc:// and without the port
		if(kvi_strEqualCIN( srv.ptr(), word, word.len()))
		{
			matches->append(new KviStr(srv));
		}

	}
}

int KviConsole::getCommonChannels(const char * nick,KviStr &chansBuffer,bool bEscapeSeq)
{
	if(!m_pConnectionInfo)return 0;
	int count = 0;
	for(KviChannel * c = m_pConnectionInfo->pChannelList->first();c;c = m_pConnectionInfo->pChannelList->next())
	{
		if(c->isOn(nick))
		{
			if(chansBuffer.hasData())chansBuffer.append(", ");
			char uFlag = c->getUserFlag(nick);
			if(uFlag)
			{
				chansBuffer.append(KviStr::Format,bEscapeSeq ? "%c\r!c\r%s\r" : "%c%s",uFlag,c->name());
			} else {
				if(bEscapeSeq)chansBuffer.append(KviStr::Format,"\r!c\r%s\r",c->name());
				else chansBuffer.append(c->name());
			}
			count ++;
		}
	}
	return count;
}


void KviConsole::getUserTipText(const char * nick,KviIrcUserEntry *e,QString &buffer)
{
	buffer = "<table width=\"100%\">" \
				"<tr><td bgcolor=\"#303030\">" \
					"<center><font color=\"#FFFFFF\"><b>";

	buffer += nick;
	buffer += "!";
	buffer += *(e->user()) ? e->user() : "*";
	buffer += "@";
	buffer += *(e->host()) ? e->host() : "*";


	buffer += "</b></font></center></td></tr>";

	if(e->avatar())
	{
		//debug("DEFAULT FACTORY: %d",QMimeSourceFactory::defaultFactory());
		QMimeSourceFactory::defaultFactory()->setPixmap("ulv_avatar",*(e->avatar()->pixmap()));
		//QMimeSourceFactory::defaultFactory()->setImage("ulv_avatar",e->avatar()->pixmap()->convertToImage());
		buffer += "<tr><td><center><img src=\"ulv_avatar\"></center></td></tr>";
	}

	if(e->hasRealName())
	{
		buffer += "<tr><td><center><b>";
		buffer += e->realName();
		buffer += "</b></center></td></tr>";
	}

	if(*(e->user()))
	{
		KviRegisteredMask *u = g_pRegisteredUserDataBase->findMatchingMask(nick,
			e->user(),e->host());
		if(u)
		{
			KviStr mask;
			u->mask()->mask(mask);
			buffer += "<tr bgcolor=\"#F0F0F0\"><td>";
			buffer += __tr2qs("Registered as");
			buffer += " <b>";
			buffer += u->user()->name();
			buffer += "</b> <font size=\"-1\">(";
			buffer += __tr2qs("Matched by");
			buffer += " ";
			buffer += mask.ptr();
			buffer += ")</td></tr>";
		}
	}

	KviStr chans;

	if(getCommonChannels(nick,chans,false))
	{
		buffer += "<tr><td bgcolor=\"#F0F0F0\">";
		buffer += __tr2qs("On <b>");
		buffer += chans.ptr();
		buffer += "</b></td></tr>";
	}

	if(e->hasServer())
	{
		buffer += "<tr><td bgcolor=\"#F0F0F0\"><nobr>";
		buffer += __tr2qs("Using server <b>%1</b>").arg(e->server());

		if(e->hasHops())
		{
			buffer += " (";
			buffer += __tr2qs("%1 hops").arg(e->hops());
			buffer += ")</nobr></td></tr>";
		} else {
			buffer += "</nobr></td></tr></table>";
		}
	}
}


void KviConsole::toggleNotifyView()
{
	if(m_pNotifyListView->isVisible())
	{
		m_pNotifyListView->hide();
		if(m_pNotifyViewButton->isOn())m_pNotifyViewButton->setOn(false);
	} else {
		m_pNotifyListView->show();
		if(!(m_pNotifyViewButton->isOn()))m_pNotifyViewButton->setOn(true);
	}
}

void KviConsole::executeInternalCommand(int index)
{
	g_pUserParser->parseCommandBuffer(kvi_getInternalCommandBuffer(index),this);
}

void KviConsole::saveProperties(KviConfig *cfg)
{
	KviWindow::saveProperties(cfg);
	cfg->writeEntry("Splitter",m_pSplitter->sizes());
}

void KviConsole::getBaseLogFileName(KviStr &buffer)
{
	buffer.sprintf("CONSOLE%u",ircContextId());
}

void KviConsole::loadProperties(KviConfig *cfg)
{
	int w = width();
	QValueList<int> def;
	def.append((w * 85) / 100);
	def.append((w * 15) / 100);
	m_pSplitter->setSizes(cfg->readIntListEntry("Splitter",def));
	KviWindow::loadProperties(cfg);
}

void KviConsole::textViewRightClicked()
{
	TRIGGER_EVENT(KviEvent_OnConsolePopupRequest,this);
}

KviWhoisInfo * KviConsole::lookupAsyncWhois(const char * nick)
{
	if(m_pConnectionInfo->pWhoisInfoList->isEmpty())return 0;
	for(KviWhoisInfo * i = m_pConnectionInfo->pWhoisInfoList->first();i;i = m_pConnectionInfo->pWhoisInfoList->next())
	{
		if(kvi_strEqualCI(nick,i->szNick.ptr()))return i;
	}
	return 0;
}

void KviConsole::partAllChannels()
{
	if(!m_pConnectionInfo)return;
	for(KviChannel * c = m_pConnectionInfo->pChannelList->first();c;c = m_pConnectionInfo->pChannelList->next())
	{
		c->close();
	}
}

void KviConsole::closeAllChannels()
{
	if(!m_pConnectionInfo)return;
	while(m_pConnectionInfo->pChannelList->first())
	{
		m_pFrm->closeWindow(m_pConnectionInfo->pChannelList->first());
	}
}

void KviConsole::closeAllQueries()
{
	if(!m_pConnectionInfo)return;
	while(m_pConnectionInfo->pQueryList->first())
	{
		m_pFrm->closeWindow(m_pConnectionInfo->pQueryList->first());
	}
}

KviWindow * KviConsole::activeWindow()
{
	if(g_pApp->activeWindow())
	{
		if(g_pApp->activeWindow()->console() == this)return g_pApp->activeWindow();
	}
	return this;
}


void KviConsole::startNotifyList()
{
	// start the notify list in 3 seconds
	// We have this delay to wait an eventual RPL_PROTOCTL from the server
	// telling us that the WATCH notify list method is supported
	__range_invalid(m_pConnectionInfo->pNotifyListTimer);

	m_pConnectionInfo->pNotifyListTimer = new QTimer();
	connect(m_pConnectionInfo->pNotifyListTimer,SIGNAL(timeout()),this,SLOT(restartNotifyList()));
	m_pConnectionInfo->pNotifyListTimer->start(15000,true);

	// This delay is large enough to fire after the MOTD has been sent,
	// even on the weirdest network.
	// If there is no MOTD, this timer will fire after 15 secs,
	// If there is a MOTD , restartNotifyList() will be triggered by RPL_ENDOFMOTD and
	// will kill the timer before it has fired.
}


void KviConsole::restartNotifyList()
{
	if(!m_pConnectionInfo)return;

	if(m_pConnectionInfo->pNotifyListTimer)
	{
		delete m_pConnectionInfo->pNotifyListTimer;
		m_pConnectionInfo->pNotifyListTimer = 0;
	}

	// clear it
	if(m_pConnectionInfo->pNotifyListManager)
	{
		m_pConnectionInfo->pNotifyListManager->stop(); // may need to remove watch entries
		delete m_pConnectionInfo->pNotifyListManager;
		m_pConnectionInfo->pNotifyListManager = 0;
	}

	if(!KVI_OPTION_BOOL(KviOption_boolUseNotifyList))return;

	if(m_pConnectionInfo->bServerSupportsWatchList && KVI_OPTION_BOOL(KviOption_boolUseWatchListIfAvailable))
	{
		if(KVI_OPTION_BOOL(KviOption_boolBeVerbose))
		{
			output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("The server seems to support the WATCH notify list method, will try to use it"));
		}
		m_pConnectionInfo->pNotifyListManager = new KviWatchNotifyListManager(this);
	} else {
		if(KVI_OPTION_BOOL(KviOption_boolUseIntelligentNotifyListManager))
		{
			m_pConnectionInfo->pNotifyListManager = new KviIsOnNotifyListManager(this);
		} else {
			m_pConnectionInfo->pNotifyListManager = new KviStupidNotifyListManager(this);
		}
	}
	m_pConnectionInfo->pNotifyListManager->start();
}

//KviWindow * KviConsole::activeWindow()
//{
//#warning "FIX THIS....ACTIVE WINDOW ?"
//	return this;
//}

KviChannel * KviConsole::createChannel(const char * szName)
{
	__range_valid(m_pConnectionInfo);
	KviChannel * c = m_pIrcContext->findDeadChannel(szName);
	if(c)
	{
		c->setAliveChan();
		c->raise();
		c->setFocus();
	} else {
		c = new KviChannel(m_pFrm,this,szName);
		m_pFrm->addWindow(c);
	}
	return c;
}

KviQuery * KviConsole::createQuery(const char * szName)
{
	__range_valid(m_pConnectionInfo);
	KviQuery * q = m_pIrcContext->findDeadQuery(szName);
	if(!q)q = findQueryByName(szName);
	if(q)
	{
		q->setAliveQuery();
		if(!KVI_OPTION_BOOL(KviOption_boolCreateMinimizedQuery))
		{
			q->raise();
			q->setFocus();
		}
	} else {
		q = new KviQuery(m_pFrm,this,szName);
		m_pFrm->addWindow(q,!KVI_OPTION_BOOL(KviOption_boolCreateMinimizedQuery));
		if(KVI_OPTION_BOOL(KviOption_boolCreateMinimizedQuery))q->minimize();
	}
	return q;
}



//KviQuery * KviConsole::findQuery(const char * nick)
//{
//	if(!connection())return 0;
//	for(KviQuery * c = connection()->pQueryList->first();c;c = connection()->pQueryList->next())
//	{
//		if(c->haveTarget(nick))return c;
//	}
//	return 0;
//}

KviQuery * KviConsole::findQueryByName(const char * name)
{
	if(!connection())return 0;
	for(KviQuery * c = connection()->pQueryList->first();c;c = connection()->pQueryList->next())
	{
		if(kvi_strEqualCI(name,c->name()))return c;
	}
	return 0;
}

void KviConsole::registerChannel(KviChannel * c)
{
	__range_valid(m_pConnectionInfo);
	m_pConnectionInfo->pChannelList->append(c);
	g_pApp->addRecentChannel(c->name());
}

void KviConsole::setChannelAsDead(KviChannel * c)
{
	if(!m_pConnectionInfo->pChannelList->removeRef(c))
		debug("Ops... setting as dead a channel that wasn't alive");
	m_pIrcContext->registerDeadChannel(c);
}

void KviConsole::setQueryAsDead(KviQuery * q)
{
	if(!m_pConnectionInfo->pQueryList->removeRef(q))
		debug("Ops... setting as dead a query that wasn't alive");
	m_pIrcContext->registerDeadQuery(q);
}

void KviConsole::setChannelAsAlive(KviChannel * c)
{
	if(!m_pIrcContext->unregisterDeadChannel(c))
		debug("Ops... setting as alive a channel that wasn't dead");
	m_pConnectionInfo->pChannelList->append(c);
}

void KviConsole::setQueryAsAlive(KviQuery * q)
{
	if(!m_pIrcContext->unregisterDeadQuery(q))
		debug("Ops... setting as alive a query that wasn't dead");
	m_pConnectionInfo->pQueryList->append(q);
}

void KviConsole::keepChannelsOpenAfterDisconnect()
{
	if(!connection())return;
	while(KviChannel * c = connection()->pChannelList->first())
	{
		c->outputNoFmt(KVI_OUT_SOCKETERROR,__tr2wc("Connection to server lost"));
		c->setDeadChan();
	}
}

void KviConsole::keepQueriesOpenAfterDisconnect()
{
	if(!connection())return;
	while(KviQuery * q = connection()->pQueryList->first())
	{
		q->outputNoFmt(KVI_OUT_SOCKETERROR,__tr2wc("Connection to server lost"));
		q->setDeadQuery();
	}
}

void KviConsole::resurrectDeadQueries()
{
	while(KviQuery * q = m_pIrcContext->firstDeadQuery())
	{
		q->outputNoFmt(KVI_OUT_SOCKETMESSAGE,__tr2wc("Connection to server established"));
		q->setAliveQuery();
	}
}

void KviConsole::unregisterChannel(KviChannel * c)
{
	if(m_pConnectionInfo)
	{		
		if(m_pConnectionInfo->pChannelList->removeRef(c))return;
	}

	// dead chan ?
	if(!m_pIrcContext->unregisterDeadChannel(c))debug("Ops... unregistering a non registered channel ?");
}

void KviConsole::registerQuery(KviQuery * c)
{
	__range_valid(m_pConnectionInfo);
	m_pConnectionInfo->pQueryList->append(c);
}


void KviConsole::unregisterQuery(KviQuery * c)
{
	if(m_pConnectionInfo)
	{
		if(m_pConnectionInfo->pQueryList->removeRef(c))return;
	}
	// may be only a dead chan
	if(!m_pIrcContext->unregisterDeadQuery(c))debug("Ops... unregistering a non registered query?");
}

QPixmap * KviConsole::myIconPtr()
{
	return g_pIconManager->getSmallIcon(KVI_SMALLICON_CONSOLE);
}


void KviConsole::getTaskBarTipText(QString &buffer)
{
	fillStatusString();

	static QString html_bold("<b>");
	static QString html_tab("&nbsp;&nbsp;");
	static QString html_eofbold("</b>");
	static QString html_hrbr("<br><hr>");
	static QString html_channel(__tr2qs("channel"));
	static QString html_channels(__tr2qs("channels"));
	static QString html_query(__tr2qs("query"));
	static QString html_queries(__tr2qs("queries"));
	static QString html_space(" ");
	static QString html_commaspace(", ");
	static QString html_br("<br>");
	static QString html_spaceparopen(" (");
	static QString html_spaceparclosed(")");


	buffer = "<table width=\"100%\">" \
						"<tr><td bgcolor=\"#303030\">" \
							"<font color=\"#FFFFFF\"><b>";
	buffer += m_szStatusString;
	buffer += "</b></font></td></tr>";

	if((m_state == Connected) && connectionInfo())
	{
		QString num;

		unsigned int uD;
		unsigned int uH;
		unsigned int uM;
		unsigned int uS;

		uD = m_pConnectionInfo->pChannelList->count();
		uH = m_pConnectionInfo->pQueryList->count();

		if(uD || uH > 0)
		{
			buffer += "<tr><td>";
			buffer += html_tab;

			if(uD > 0)
			{
				num.setNum(uD);

				buffer += html_bold;
				buffer += num;
				buffer += html_eofbold;
				buffer += html_space;
				buffer += uD > 1 ? html_channels : html_channel;
				if(uH > 0)
					buffer += html_commaspace;
			}

			if(uH > 0)
			{
				num.setNum(uH);
				buffer += html_bold;
				buffer += num;
				buffer += html_eofbold;
				buffer += html_space;
				buffer += uH > 1 ? html_queries : html_query;
			}
			buffer += "</td></tr>";
		}

		kvi_secondsToDaysHoursMinsSecs(
			(unsigned int)(kvi_secondsSince(m_pConnectionInfo->tConnectionStart)),
			&uD,&uH,&uM,&uS);

		QDateTime dt;
		dt.setTime_t(m_pConnectionInfo->tConnectionStart);

		buffer += "<tr><td bgcolor=\"#F0F0F0\">";

		buffer += __tr2qs("Connected since");
		buffer += html_space;
		buffer += html_br;
		buffer += html_tab;
		buffer += html_bold;
		buffer += dt.toString();
		buffer += html_eofbold;
		buffer += html_br;

		//buffer += html_spaceparopen;

		KviWStr tspan;

		tspan.sprintf(__tr2wc("%ud %uh %um %us"),uD,uH,uM,uS);

		buffer += __tr2qs("Online for");
		buffer += html_space;
		buffer += html_bold;
		buffer += tspan.qstring();
		buffer += html_eofbold;
		//buffer += html_spaceparclosed;

		buffer += "</td></tr><tr><td bgcolor=\"#F0F0F0\">";

		kvi_secondsToDaysHoursMinsSecs(
			(unsigned int)(kvi_secondsSince(m_pConnectionInfo->tLastMessage)),
			&uD,&uH,&uM,&uS);

		dt.setTime_t(m_pConnectionInfo->tLastMessage);

		buffer += __tr2qs("Server idle for");
		buffer += html_space;
		buffer += html_bold;
		tspan.sprintf(__tr2wc("%ud %uh %um %us"),uD,uH,uM,uS);
		buffer += tspan.qstring();
		buffer += "</b></td></tr>";
	}

	buffer += "</table>";


}

void KviConsole::fillStatusString()
{
	switch(m_state)
	{
		case NotConnected: m_szStatusString = __tr2qs("No connection"); break;
		case Connecting:   m_szStatusString = __tr2qs("Connection in progress..."); break;
		case LoggingIn:    m_szStatusString = __tr2qs("Login in progress..."); break;
		case Connected:
			m_szStatusString = currentNickName();
			if(m_pConnectionInfo->szUserMode.hasData())
			{
				m_szStatusString += " (+";
				m_szStatusString += m_pConnectionInfo->szUserMode.ptr();

				if(m_pConnectionInfo->bAway)
				{
					m_szStatusString += QChar(' ');
					m_szStatusString += __tr2qs("away");
				}
				m_szStatusString += QChar(')');
			} else {
				if(m_pConnectionInfo->bAway)
				{
					m_szStatusString += " (";
					m_szStatusString += __tr2qs("away");
					m_szStatusString += QChar(')');
				}
			}

			m_szStatusString += __tr2qs(" on ");
			m_szStatusString += currentServerName();
		break;
	}
}

void KviConsole::fillCaptionBuffers()
{
	fillStatusString();

	static QString begin("<nobr><font color=\"");
	static QString boldbegin("\"><b>");
	static QString endofbold("</b></font> <font color=\"");
	static QString endoffont("\">");
	static QString end("</font></nobr>");

	m_szPlainTextCaption = name();
	m_szPlainTextCaption += " [";
	m_szPlainTextCaption += m_szStatusString;
	m_szPlainTextCaption += QChar(']');

	m_szHtmlActiveCaption = begin;
	m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive).name();
	m_szHtmlActiveCaption += boldbegin;
	m_szHtmlActiveCaption += name();
	m_szHtmlActiveCaption += endofbold;
	m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive2).name();
	m_szHtmlActiveCaption += endoffont;
	m_szHtmlActiveCaption += m_szStatusString;
	m_szHtmlActiveCaption += end;

	m_szHtmlInactiveCaption = begin;
	m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive).name();
	m_szHtmlInactiveCaption += boldbegin;
	m_szHtmlInactiveCaption += name();
	m_szHtmlInactiveCaption += endofbold;
	m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive2).name();
	m_szHtmlInactiveCaption += endoffont;
	m_szHtmlInactiveCaption += m_szStatusString;
	m_szHtmlInactiveCaption += end;
}

bool KviConsole::setUserMode(char mode,bool bSet)
{
	__range_valid(m_pConnectionInfo);
	if(bSet)
	{
		if(m_pConnectionInfo->szUserMode.contains(mode))return false;
		m_pConnectionInfo->szUserMode.append(mode);
	} else {
		if(m_pConnectionInfo->szUserMode.occurences(mode) < 1)return false;
		KviStr m = mode;
		m_pConnectionInfo->szUserMode.replaceAll(m.ptr(),"");
	}
	updateCaption();
	m_pFrm->ircContextToolBar()->updateStatusDisplay(this);
//	m_pIcController->m_pStatusLabel->update();
	return true;
}

void KviConsole::changeAwayState(bool bAway)
{
	__range_valid(m_pConnectionInfo);
	m_pConnectionInfo->bAway = bAway;
	if(bAway)m_pConnectionInfo->tAwayTime = kvi_unixTime();
	updateCaption();
	m_pFrm->ircContextToolBar()->updateStatusDisplay(this);
}

void KviConsole::resizeEvent(QResizeEvent *e)
{
	int hght = m_pInput->heightHint();
	int hght2 = m_pButtonBox->sizeHint().height();
	m_pButtonBox->setGeometry(0,0,width(),hght2);
	m_pSplitter->setGeometry(0,hght2,width(),height() - (hght + hght2));
	m_pInput->setGeometry(0,height() - hght,width(),hght);
}

QSize KviConsole::sizeHint() const
{
	QSize ret(m_pIrcView->sizeHint().height(),m_pIrcView->sizeHint().height() + m_pInput->heightHint());
	return ret;
}

void KviConsole::asyncServerCommand()
{
	if(m_pConnectionInfo)
	{
		// need a brutal disconnect here
		if(m_state == Connected)m_pLink->sendFmtData("QUIT :Changing server...");
		m_pConnectionInfo->bSentQuit = true;
		connectButtonClicked();
	}
	connectToCurrentServer();
}

void KviConsole::terminateConnectionRequest(bool bForce,const char * quitMsg)
{
	if(!m_pConnectionInfo)return; // hm ?

	switch(m_state)
	{
		case Connected:
		{
			// was connected
			KviStr szQuitMsg = quitMsg ? quitMsg : "";
			if(szQuitMsg.isEmpty())szQuitMsg = KVI_OPTION_STRING(KviOption_stringQuitMessage);
			m_pLink->sendFmtData("QUIT :%s",szQuitMsg.ptr());

			if(KVI_OPTION_BOOL(KviOption_boolForceBrutalQuit) || m_pConnectionInfo->bSentQuit || bForce)
			{
				m_pConnectionInfo->bSentQuit = true;
				m_pLink->abortConnection();
			} else {
				m_pConnectionInfo->bSentQuit = true;
				outputNoFmt(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Sent QUIT, waiting for the server to close the connection..."));
			}
		}
		break;
		case Connecting:
		{
			if(m_pConnectionInfo->pDns)
			{
				// was waiting for dns... kill the dns and die
				outputNoFmt(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Connection aborted"));
				setState(NotConnected);
			} else {
				// was waiting for connection to estabilish, abort it
				m_pLink->abortConnection();
			}
		}
		break;
		case LoggingIn:
			// was waiting for login, just abort it
			m_pLink->abortConnection();
		break;
		default:
			debug("Opss.. terminateConnectionRequest() when there is no connection");
		break;
	}
}

void KviConsole::connectButtonClicked()
{
	if(m_pConnectionInfo == 0)
	{
		// No connections in progress
		connectToCurrentServer();
	} else {
		// Sth is going on
		terminateConnectionRequest(false);
	}
}

void KviConsole::setState(ConnectionState st)
{
	if(m_state != st)
	{
		m_state = st;
		switch(st)
		{
			case NotConnected:
				m_pFrm->ircContextToolBar()->connectStateChanged(this);
				destroyConnection();
			break;
			case LoggingIn:
			case Connecting:
			case Connected:
				m_pFrm->ircContextToolBar()->connectStateChanged(this);
			break;
		}
		updateCaption();
		emit stateChange(m_state);
	}
}

void KviConsole::setUserInfoFromServer(const char * username,const char *localhost)
{
	__range_valid(m_pConnectionInfo);
	m_pConnectionInfo->szUserNameFromServer = username;
	if(localhost && *localhost && !kvi_strEqualCS(m_pConnectionInfo->szLocalHostFromServer.ptr(),localhost))
	{
		static bool warned_once = false;

		if(!warned_once)
		{
			if(!(m_pConnectionInfo->szLocalHostFromServer.isEmpty() ||
#ifdef COMPILE_IPV6_SUPPORT
				kvi_isValidStringIp(m_pConnectionInfo->szLocalHostFromServer.ptr()) ||
				kvi_isValidStringIp_V6(m_pConnectionInfo->szLocalHostFromServer.ptr())))
#else
				kvi_isValidStringIp(m_pConnectionInfo->szLocalHostFromServer.ptr())))
#endif
			{
				output(KVI_OUT_SYSTEMWARNING,__tr2wc("The server seems to have changed the idea about the local hostname"));
				output(KVI_OUT_SYSTEMWARNING,__tr2wc("You're probably using a broken bouncer or maybe something weird is happening on the IRC server"));
				warned_once = true;
			}
		}

		m_pConnectionInfo->szLocalHostFromServer = localhost;

		bool bChangeIp = true;

#ifdef COMPILE_IPV6_SUPPORT
		if((kvi_isValidStringIp(m_pConnectionInfo->szLocalIpFromServer.ptr()) && kvi_isRoutableIpString(m_pConnectionInfo->szLocalIpFromServer.ptr())) ||
			kvi_isValidStringIp_V6(m_pConnectionInfo->szLocalIpFromServer.ptr()))
#else
		if((kvi_isValidStringIp(m_pConnectionInfo->szLocalIpFromServer.ptr()) && kvi_isRoutableIpString(m_pConnectionInfo->szLocalIpFromServer.ptr())))
#endif
		{
			if(KVI_OPTION_BOOL(KviOption_boolDccGuessIpFromServerWhenLocalIsUnroutable) && KVI_OPTION_BOOL(KviOption_boolDccBrokenBouncerHack))
			{
				outputNoFmt(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Here goes your \"broken bouncer hack\": The server has changed the hostname but I'll ignore the IP address change"));
				bChangeIp = false;
			}
		}

		if(bChangeIp)
		{
#ifdef COMPILE_IPV6_SUPPORT
			if(kvi_isValidStringIp(localhost) || kvi_isValidStringIp_V6(localhost))
#else
			if(kvi_isValidStringIp(localhost))
#endif
			{
				output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("The local IP address as seen by the IRC server is %s"),localhost);
				m_pConnectionInfo->szLocalIpFromServer = m_pConnectionInfo->szLocalHostFromServer;
			} else {
				// look it up too
				if(m_pConnectionInfo->pLocalhostDns)delete m_pConnectionInfo->pLocalhostDns; // it could be only another local host lookup
				m_pConnectionInfo->pLocalhostDns = new KviDns();
				connect(m_pConnectionInfo->pLocalhostDns,SIGNAL(lookupDone(KviDns *)),this,
					SLOT(localHostLookupTerminated(KviDns *)));

				if(!m_pConnectionInfo->pLocalhostDns->lookup(m_pConnectionInfo->szLocalHostFromServer.ptr(),KviDns::Any))
				{
					outputNoFmt(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Unable to resolve the local hostname as seen by the IRC server: Hostname is masked?"));
					delete m_pConnectionInfo->pLocalhostDns;
					m_pConnectionInfo->pLocalhostDns = 0;
				} else {
					outputNoFmt(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Looking up the local hostname as seen by the IRC server..."));
				}
			}
		}
	}
}


void KviConsole::localHostLookupTerminated(KviDns *)
{
	if(!m_pConnectionInfo)return;
	if(!m_pConnectionInfo->pLocalhostDns)return;

	if(m_pConnectionInfo->pLocalhostDns->state() != KviDns::Success)
	{
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Unable to resolve the local hostname as seen by the IRC server: %s"),
			kvi_getErrorString(m_pConnectionInfo->pLocalhostDns->error()));
	} else {
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Local hostname as seen by the IRC server resolved to %s"),m_pConnectionInfo->pLocalhostDns->firstIpAddress());
		m_pConnectionInfo->szLocalIpFromServer = m_pConnectionInfo->pLocalhostDns->firstIpAddress();
	}

	delete m_pConnectionInfo->pLocalhostDns;
	m_pConnectionInfo->pLocalhostDns = 0;
}

void KviConsole::nickChange(const char * newNick)
{
	__range_valid(m_pConnectionInfo);
	m_pConnectionInfo->szNickName = newNick;
	output(KVI_OUT_NICK,__tr2wc("You have changed your nickname to %s"),newNick);
	updateCaption();
	m_pFrm->ircContextToolBar()->updateStatusDisplay(this);
	g_pApp->addRecentNickname(newNick);
}

void KviConsole::setServerInfoFromServer(const char * servname,const char * umodes,const char * chanmodes)
{
	__range_valid(m_pConnectionInfo);
	if(servname)m_pConnectionInfo->szServerNameFromServer = servname;
	if(umodes)m_pConnectionInfo->szServerSupportedUserModes = umodes;
	if(chanmodes)
	{
		m_pConnectionInfo->szServerSupportedChanModes = chanmodes;
		m_pConnectionInfo->bServerSupportsModeIe = (m_pConnectionInfo->szServerSupportedChanModes.contains('e') &&
			m_pConnectionInfo->szServerSupportedChanModes.contains('I'));
	}
	updateCaption(); // for server name
	m_pFrm->ircContextToolBar()->updateStatusDisplay(this);
}

void KviConsole::setAsyncServerCommandData(KviAsyncServerCommandData * d)
{
	destroyAsyncServerCommandData();
	m_pAsyncServerCommandData = d;
}

void KviConsole::destroyAsyncServerCommandData()
{
	if(m_pAsyncServerCommandData)
	{
		delete m_pAsyncServerCommandData;
		m_pAsyncServerCommandData = 0;
	}
}



void KviConsole::createLink(const KviStr &szLinkFilterName)
{
	if(m_pLink)destroyLink();

	if(szLinkFilterName.isEmpty())goto alloc_builtin_link;
	if(szLinkFilterName.equalsCI("irc"))goto alloc_builtin_link;

	m_pLink = (KviIrcLink *)g_pModuleExtensionManager->allocateExtension("linkfilter",szLinkFilterName.ptr(),this,0,0,szLinkFilterName.ptr());

	if(m_pLink)
	{
		connect(m_pLink,SIGNAL(destroyed()),this,SLOT(linkDestroyed()));
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Using filtered IRC protocol: Link filter is \"%s\""),szLinkFilterName.ptr());
		return;
	}

	output(KVI_OUT_SYSTEMWARNING,__tr2wc("Failed to set up the link filter \"%s\", will try with plain IRC"),szLinkFilterName.ptr());

alloc_builtin_link:
	m_pLink = new KviIrcLink(this);
}

void KviConsole::linkDestroyed()
{
	// ops.. the link has been destroyed without permission :D
	m_pLink = 0;
	output(KVI_OUT_SYSTEMWARNING,__tr2wc("Ops... for some reason the link object has been destroyed"));
	createLink(KviStr::emptyString()); // we need a new one
}

void KviConsole::destroyLink()
{
	if(!m_pLink)return;
	// disconnect , in particular kill the destroyed() signal
	disconnect(m_pLink,0,this,0);
	// the module extension server links must be destroyed in the module that provided it
	if(m_pLink->inherits("KviMexLinkFilter"))((KviMexLinkFilter *)m_pLink)->die();
	else delete m_pLink; // we can simply delete the builtin objects
	m_pLink = 0;
}

void KviConsole::connectToCurrentServer()
{
	__range_invalid(m_pConnectionInfo);

	outputNoFmt(KVI_OUT_SYSTEMMESSAGE,""); // spacer

	bool bGotServer = false;

	if(m_pAsyncServerCommandData)
	{
		if(m_pAsyncServerCommandData->szServer.hasData())
		{
			KviIrcServerDefinition d;
			d.szServer = m_pAsyncServerCommandData->szServer;
			d.bPortIsValid = m_pAsyncServerCommandData->bPortIsOk;
			d.uPort = m_pAsyncServerCommandData->uPort;
			d.bIpV6 = m_pAsyncServerCommandData->bUseIpV6;
			d.bSSL = m_pAsyncServerCommandData->bUseSSL;
			d.szLinkFilter = m_pAsyncServerCommandData->szLinkFilter;
			d.szPass = m_pAsyncServerCommandData->szPass;
			d.szNick = m_pAsyncServerCommandData->szNick;
			d.szInitUMode = m_pAsyncServerCommandData->szInitUMode;
			g_pIrcServerDataBase->makeCurrentServer(&d);
			bGotServer = true;
		}
	}

	/*

	This breaks the servers dialog: one can't select a different server
	since the last is always used.
	It needs a more strict condition or, better yet, an asyncServerCommandData
	set when this console has lost the connection.

	if(!bGotServer)
	{
		// try to force the last server this console has used
		if(m_pLastIrcServer)
		{
			KviIrcServerDefinition d;
			d.szServer = m_pLastIrcServer->m_szHostname;
			d.bPortIsValid = true;
			d.uPort = m_pLastIrcServer->port();
			d.bIpV6 = m_pLastIrcServer->isIpV6();
			d.bSSL = m_pLastIrcServer->useSSL();
			d.szLinkFilter = m_pLastIrcServer->linkFilter();
			d.szPass = m_pLastIrcServer->password();
			d.szNick = m_pLastIrcServer->nickname();
			d.szInitUMode = m_pLastIrcServer->m_szInitUMode;
			g_pIrcServerDataBase->makeCurrentServer(&d);
			bGotServer = true;
		}
	}

	*/

	KviIrcNetwork * net = g_pIrcServerDataBase->currentNetwork();
	KviIrcServer  * srv = 0;
	KviProxy      * prx = 0;

	if(net)srv = net->currentServer();

	if(!srv)
	{
		outputNoFmt(KVI_OUT_SYSTEMERROR,__tr2wc("No servers available. Check the options dialog or use the /SERVER command"));
		destroyAsyncServerCommandData();
		return;
	}

	if(KVI_OPTION_BOOL(KviOption_boolUseProxyHost))
	{
		prx = g_pProxyDataBase->currentProxy();
		if(!prx)
		{
			outputNoFmt(KVI_OUT_SYSTEMWARNING,__tr2wc("No proxy hosts available, resuming direct connection"));
		}
	}

	m_pConnectionInfo = new KviConnectionInfo(net,srv,prx);


	destroyLink();
	createLink(srv->linkFilter());

	if(m_pAsyncServerCommandData)
	{
		m_pConnectionInfo->szCommandToExecAfterConnect = m_pAsyncServerCommandData->szCommandToExecAfterConnect;
		m_pConnectionInfo->szBindAddress               = m_pAsyncServerCommandData->szBindAddress;
		destroyAsyncServerCommandData();
	}

	m_pNotifyListView->setUserDataBase(m_pConnectionInfo->pUserDataBase);

	setState(Connecting);

	output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Attempting %w to %s (%s) on port %u"),
		m_pConnectionInfo->pIrcServer->useSSL() ? __tr2wc("secure connection") : __tr2wc("connection"),
		m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
		m_pConnectionInfo->szNetworkName.ptr(),
		m_pConnectionInfo->pIrcServer->m_uPort);

	if(prx)
	{
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Attempting to 'bounce' on proxy %s on port %u (protocol %s)"),
			m_pConnectionInfo->pProxy->m_szHostname.ptr(),
			m_pConnectionInfo->pProxy->m_uPort,
			m_pConnectionInfo->pProxy->protocolName());
		lookupProxyHostname();
	} else {
		lookupServerHostname();
	}
}

void KviConsole::lookupProxyHostname()
{
	bool bValidIp;

#ifdef COMPILE_IPV6_SUPPORT
	if(m_pConnectionInfo->pProxy->isIpV6())
	{
		bValidIp = kvi_isValidStringIp_V6(m_pConnectionInfo->pProxy->m_szIp.ptr());
	} else {
#endif
		bValidIp = kvi_isValidStringIp(m_pConnectionInfo->pProxy->m_szIp.ptr());
#ifdef COMPILE_IPV6_SUPPORT
	}
#endif

// FIXME: #warning "option for caching proxy ip addresses too ?"
// FIXME: #warning "test IPv6 proxies ?"

	if(bValidIp)
	{
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Using cached proxy IP address (%s)"),
			m_pConnectionInfo->pProxy->m_szIp.ptr());
		lookupServerHostname();
	} else {
		if(m_pConnectionInfo->pDns)
		{
			debug("Something weird is happening, pDns is non-zero in lookupProxyHostname()");
			delete m_pConnectionInfo->pDns;
		}
		m_pConnectionInfo->pDns = new KviDns();
		connect(m_pConnectionInfo->pDns,SIGNAL(lookupDone(KviDns *)),this,
			SLOT(proxyLookupTerminated(KviDns *)));

		if(!m_pConnectionInfo->pDns->lookup(m_pConnectionInfo->pProxy->m_szHostname.ptr(),
			m_pConnectionInfo->pProxy->isIpV6() ? KviDns::IpV6 : KviDns::IpV4))
		{
			outputNoFmt(KVI_OUT_SYSTEMWARNING,__tr2wc("Unable to look up the IRC proxy hostname: Can't start the DNS slave"));
			outputNoFmt(KVI_OUT_SYSTEMWARNING,__tr2wc("Resuming direct server connection"));
// FIXME: #warning "Option for resuming direct connection or not ?"
			delete m_pConnectionInfo->pDns;
			m_pConnectionInfo->pDns = 0;
			delete m_pConnectionInfo->pProxy;
			m_pConnectionInfo->pProxy = 0;
			lookupServerHostname();
		} else {
			output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Looking up the proxy hostname (%s)..."),
				m_pConnectionInfo->pProxy->m_szHostname.ptr());
		}
	}
}


void KviConsole::lookupServerHostname()
{
	bool bValidIp;

#ifdef COMPILE_IPV6_SUPPORT
	if(m_pConnectionInfo->pIrcServer->isIpV6())
	{
		bValidIp = kvi_isValidStringIp_V6(m_pConnectionInfo->pIrcServer->m_szIp.latin1());
	} else {
#endif
		bValidIp = kvi_isValidStringIp(m_pConnectionInfo->pIrcServer->m_szIp.latin1());
#ifdef COMPILE_IPV6_SUPPORT
	}
#endif

	if(bValidIp && m_pConnectionInfo->pIrcServer->cacheIp())
	{
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Using cached server IP address (%s)"),
			m_pConnectionInfo->pIrcServer->m_szIp.latin1());
		haveServerIp();
	} else {
		if(m_pConnectionInfo->pDns)
		{
			debug("Something weird is happening, pDns is non-zero in lookupServerHostname()");
			delete m_pConnectionInfo->pDns;
		}
		m_pConnectionInfo->pDns = new KviDns();
		connect(m_pConnectionInfo->pDns,SIGNAL(lookupDone(KviDns *)),this,
			SLOT(serverLookupTerminated(KviDns *)));
		if(!m_pConnectionInfo->pDns->lookup(m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
			m_pConnectionInfo->pIrcServer->isIpV6() ? KviDns::IpV6 : KviDns::IpV4))
		{
			outputNoFmt(KVI_OUT_SYSTEMERROR,__tr2wc("Unable to look up the server hostname: Can't start the DNS slave"));
			setState(NotConnected);
		} else {
			output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Looking up the server hostname (%s)..."),
				m_pConnectionInfo->pIrcServer->m_szHostname.latin1());
		}
	}
}

void KviConsole::proxyLookupTerminated(KviDns *)
{
	if(m_pConnectionInfo->pDns->state() != KviDns::Success)
	{
		output(KVI_OUT_SYSTEMERROR,__tr2wc("Can't find the proxy IP address: %s"),
			kvi_getErrorString(m_pConnectionInfo->pDns->error()));

// FIXME: #warning "Option to resume the direct connection if proxy failed ?"

		output(KVI_OUT_SYSTEMERROR,__tr2wc("Resuming direct server connection"));
		delete m_pConnectionInfo->pProxy;
		m_pConnectionInfo->pProxy = 0;
	} else {

		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Proxy hostname resolved to %s"),m_pConnectionInfo->pDns->firstIpAddress());
		m_pConnectionInfo->pProxy->m_szIp = m_pConnectionInfo->pDns->firstIpAddress();
		g_pProxyDataBase->updateProxyIp(m_pConnectionInfo->pProxy->m_szIp.ptr(),m_pConnectionInfo->pDns->firstIpAddress());

		if(m_pConnectionInfo->pDns->hostnameCount() > 1)
		{
			KviStr szFirstHostname = m_pConnectionInfo->pDns->firstHostname();

			for(KviStr * addr = m_pConnectionInfo->pDns->hostnameList()->next();addr;addr = m_pConnectionInfo->pDns->hostnameList()->next())
			{
				output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Proxy %s has a nickname: %s"),szFirstHostname.ptr(),addr->ptr());
			}
		}
	}
	delete m_pConnectionInfo->pDns;
	m_pConnectionInfo->pDns = 0;
// FIXME: #warning "With http proxies we might even avoid lookups for server names!"
	lookupServerHostname();
}


void KviConsole::serverLookupTerminated(KviDns *)
{
	if(m_pConnectionInfo->pDns->state() != KviDns::Success)
	{
		output(KVI_OUT_SYSTEMERROR,__tr2wc("Can't find the server IP address: %s"),
			kvi_getErrorString(m_pConnectionInfo->pDns->error()));
		output(KVI_OUT_SYSTEMERROR,__tr2wc("Connection attempt failed [%s]"),
			m_pConnectionInfo->pIrcServer->m_szHostname.latin1());
#ifdef COMPILE_IPV6_SUPPORT
		if(!(m_pConnectionInfo->pIrcServer->isIpV6()))
		{
			output(KVI_OUT_SYSTEMERROR,__tr2wc("If this server is an IPv6 one, try /server -i %s"),m_pConnectionInfo->pIrcServer->m_szHostname.latin1());
		}
#endif
		setState(NotConnected);
	} else {
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Server hostname resolved to %s"),m_pConnectionInfo->pDns->firstIpAddress());
		g_pIrcServerDataBase->updateServerIp(m_pConnectionInfo->pIrcServer->m_szHostname,m_pConnectionInfo->pIrcServer->m_uPort,m_pConnectionInfo->pDns->firstIpAddress());

		if(!kvi_strEqualCI(m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),m_pConnectionInfo->pDns->firstHostname()))
		{
			output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Real hostname for %s is %s"),
				m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
				m_pConnectionInfo->pDns->firstHostname());
			m_pConnectionInfo->pIrcServer->m_szHostname = m_pConnectionInfo->pDns->firstHostname();
		}

		m_pConnectionInfo->pIrcServer->m_szIp = m_pConnectionInfo->pDns->firstIpAddress();


		if(m_pConnectionInfo->pDns->hostnameCount() > 1)
		{
			KviStr szFirstHostname = m_pConnectionInfo->pDns->firstHostname();

			for(KviStr * addr = m_pConnectionInfo->pDns->hostnameList()->next();addr;addr = m_pConnectionInfo->pDns->hostnameList()->next())
			{
				output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Server %s has a nickname: %s"),szFirstHostname.ptr(),addr->ptr());
			}
		}

		delete m_pConnectionInfo->pDns;
		m_pConnectionInfo->pDns = 0;
		haveServerIp();
	}
}

bool KviConsole::validateLocalAddress(const char * szAddress,KviStr &szBuffer)
{
	// szAddress may be an ip address or an interface name
	__range_valid(m_pConnectionInfo);
	__range_valid(m_pConnectionInfo->pIrcServer);

#ifdef COMPILE_IPV6_SUPPORT
	if(m_pConnectionInfo->pIrcServer->isIpV6())
	{
		if(kvi_isValidStringIp_V6(szAddress))
		{
			szBuffer = szAddress;
			return true;
		}
	} else {
#endif
		if(kvi_isValidStringIp(szAddress))
		{
			szBuffer = szAddress;
			return true;
		}
#ifdef COMPILE_IPV6_SUPPORT
	}
#endif

	// is it an interface name ?

	return kvi_getInterfaceAddress(szAddress,szBuffer);
}

void KviConsole::haveServerIp()
{

	if(KVI_OPTION_BOOL(KviOption_boolUseIdentService))m_pFrm->executeInternalCommand(KVI_INTERNALCOMMAND_IDENT_START);

	KviStr bindAddress;

	if(m_pConnectionInfo->szBindAddress.hasData())
	{
		if(!validateLocalAddress(m_pConnectionInfo->szBindAddress,bindAddress))
		{
			if((m_pConnectionInfo->szBindAddress.findFirstIdx('.') != -1) ||
				(m_pConnectionInfo->szBindAddress.findFirstIdx(':') != -1))
			{
				output(KVI_OUT_SYSTEMWARNING,__tr2wc("The specified bind address (%s) is not valid"),
						m_pConnectionInfo->szBindAddress.ptr());
			} else {
				output(KVI_OUT_SYSTEMWARNING,__tr2wc("The specified bind address (%s) is not valid (the interface it refers to might be down)"),
						m_pConnectionInfo->szBindAddress.ptr());
			}
			m_pConnectionInfo->szBindAddress = "";
		}
	} else {
		// the options specify a bind address ?
#ifdef COMPILE_IPV6_SUPPORT
		if(m_pConnectionInfo->pIrcServer->isIpV6())
		{
			if(KVI_OPTION_BOOL(KviOption_boolBindIrcIpV6ConnectionsToSpecifiedAddress))
			{
				if(KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).hasData())
				{
					if(!validateLocalAddress(KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).ptr(),bindAddress))
					{
						// if it is not an interface name , kill it for now and let the user correct the address
						if(KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).findFirstIdx(':') != -1)
						{
							output(KVI_OUT_SYSTEMWARNING,__tr2wc("The system-wide IPv6 bind address (%s) is not valid"),
								KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).ptr());
							KVI_OPTION_BOOL(KviOption_boolBindIrcIpV6ConnectionsToSpecifiedAddress) = false;
						} else {
							// this is an interface address: might be down
							output(KVI_OUT_SYSTEMWARNING,__tr2wc("The system-wide IPv6 bind address (%s) is not valid (the interface it refers to might be down)"),
								KVI_OPTION_STRING(KviOption_stringIpV6ConnectionBindAddress).ptr());
						}
					}
				} else {
					// empty address....kill it
					KVI_OPTION_BOOL(KviOption_boolBindIrcIpV6ConnectionsToSpecifiedAddress) = false;
				}
			}
		} else {
#endif
			if(KVI_OPTION_BOOL(KviOption_boolBindIrcIpV4ConnectionsToSpecifiedAddress))
			{
				if(KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).hasData())
				{
					if(!validateLocalAddress(KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).ptr(),bindAddress))
					{
						// if it is not an interface name , kill it for now and let the user correct the address
						if(KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).findFirstIdx(':') != -1)
						{
							output(KVI_OUT_SYSTEMWARNING,__tr2wc("The system-wide IPv4 bind address (%s) is not valid"),
								KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).ptr());
							KVI_OPTION_BOOL(KviOption_boolBindIrcIpV4ConnectionsToSpecifiedAddress) = false;
						} else {
							// this is an interface address: might be down
							output(KVI_OUT_SYSTEMWARNING,__tr2wc("The system-wide IPv4 bind address (%s) is not valid (the interface it refers to might be down)"),
								KVI_OPTION_STRING(KviOption_stringIpV4ConnectionBindAddress).ptr());
						}
					}
				} else {
					// empty address....kill it
					KVI_OPTION_BOOL(KviOption_boolBindIrcIpV4ConnectionsToSpecifiedAddress) = false;
				}
			}
#ifdef COMPILE_IPV6_SUPPORT
		}
#endif
	}

	int iErr = m_pLink->startConnection(m_pConnectionInfo->pIrcServer,m_pConnectionInfo->pProxy,
					bindAddress.hasData() ? bindAddress.ptr() : 0);
	if(iErr != KviError_success)
	{
		output(KVI_OUT_SYSTEMERROR,__tr2wc("Connection attempt failed (%s): %s"),
			m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
			kvi_getErrorString(iErr));
		setState(NotConnected);
	}
}


void KviConsole::incomingMessage(const char * message)
{
	// A message has arrived from the current server

	// First of all , notify the monitors
	if(hasMonitors())
	{
		for(KviIrcLinkMonitor *m = monitorList()->first();m;m = monitorList()->next())
		{
			m->incomingMessage(message);
		}
	}

	// set the last message time
	m_pConnectionInfo->tLastMessage = kvi_unixTime();

	// and pass it to the server parser for processing
	g_pServerParser->parseMessage(message,this);
}

void KviConsole::linkEvent(KviIrcLink::LinkEvent ev,void * arg)
{

	// FIXME: icons for SSL message , SSL error , Proxy message and Proxy error ?
	switch(ev)
	{
		case KviIrcLink::SSLCertificate:
#ifdef COMPILE_SSL_SUPPORT
			if(KVI_OPTION_BOOL(KviOption_boolBeVerbose))
				KviSSLMaster::printSSLCertificate(this,KviWStr(__tr2wc("Server X509 certificate")).qstring(),(KviSSLCertificate *)arg);
#endif
		break;
		case KviIrcLink::SSLCipherInfo:
#ifdef COMPILE_SSL_SUPPORT
			KviSSLMaster::printSSLCipherInfo(this,KviWStr (__tr2wc("Current transmission cipher")).qstring(),(KviSSLCipherInfo *)arg);
#endif
		break;
		case KviIrcLink::SSLMessage:
			output(KVI_OUT_SSL,__tr2wc("[SSL]: %s"),(char *)arg);
		break;
		case KviIrcLink::SSLError:
			output(KVI_OUT_SOCKETERROR,__tr2wc("[SSL ERROR]: %s"),(char *)arg);
		break;
		case KviIrcLink::ProxyMessage:
			output(KVI_OUT_SOCKETMESSAGE,__tr2wc("[PROXY]: %s"),(char *)arg);
		break;
		case KviIrcLink::ProxyError:
			output(KVI_OUT_SOCKETERROR,__tr2wc("[PROXY ERROR]: %s"),(char *)arg);
		break;
		case KviIrcLink::Message:
			output(KVI_OUT_SOCKETMESSAGE,__tr2wc("[SOCKET]: %s"),(char *)arg);
		break;
		case KviIrcLink::Warning:
			output(KVI_OUT_SOCKETWARNING,__tr2wc("[SOCKET WARNING]: %s"),(char *)arg);
		break;
		case KviIrcLink::Error:
			output(KVI_OUT_SOCKETERROR,__tr2wc("[SOCKET ERROR]: %s"),kvi_getErrorString((int)arg));
			if(m_state == Connecting)
			{
				if(KVI_OPTION_BOOL(KviOption_boolAutoReconnectOnUnexpectedDisconnect) &&
					(!KVI_OPTION_UINT(KviOption_uintMaxAutoReconnectAttempts) ||
					m_uConnectAttemptCount < KVI_OPTION_UINT(KviOption_uintMaxAutoReconnectAttempts)))
				{
					m_uConnectAttemptCount++;
					//FIXME: Make it possible to stop the reconnect timer before it triggers.
					//FIXME: Multiply the delay by (m_uConnectAttemptCount / 2) so later connects are less frequent. 
					QTimer::singleShot(1000 * KVI_OPTION_UINT(KviOption_uintAutoReconnectDelay),this,
							SLOT(asyncServerCommand()));
					break;
				}
				if(m_pConnectionInfo->pIrcServer->cacheIp())
				{
					if((((int)arg) == KviError_connectionTimedOut) ||
						(((int)arg) == KviError_connectionRefused) ||
						(((int)arg) == KviError_networkUnreachable) ||
						(((int)arg) == KviError_hostUnreachable))
					{
						output(KVI_OUT_SYSTEMWARNING,__tr2wc("The connection attempt failed while using a cached IP address for the current server"));
						output(KVI_OUT_SYSTEMWARNING,__tr2wc("The problem *might* be caused by an updated DNS entry"));
						output(KVI_OUT_SYSTEMWARNING,__tr2wc("Try reconnecting with caching disabled"));
					}
				}
			}
		break;
		case KviIrcLink::StateChange:
			switch((KviIrcLink::LinkState)((int)arg))
			{
				case KviIrcLink::Connected:
				{
					if(m_pMonitorList)
					{
						for(KviIrcLinkMonitor *m =m_pMonitorList->first();m;m =m_pMonitorList->next())
						{
							m->connectionInitiated();
						}
					}
					bool bStopOutput = false;
					m_pConnectionInfo->tConnectionStart = kvi_unixTime();
					m_pConnectionInfo->tLastMessage = kvi_unixTime();

					if(g_pEventManager->hasEventHandlers(KviEvent_OnIrcConnectionEstabilished))
					{
						bStopOutput = g_pUserParser->triggerEvent(KviEvent_OnIrcConnectionEstabilished,this,0);
					}
					if(!bStopOutput)
					{
						output(KVI_OUT_CONNECTION,__tr2wc("%w established [%s (%s:%u)]"),
							m_pLink->usingSSL() ? __tr2wc("Secure connection") : __tr2wc("Connection"),
							m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
							m_pConnectionInfo->pIrcServer->m_szIp.latin1(),
							m_pConnectionInfo->pIrcServer->m_uPort);
					}

					// Add to recent server list (build the url of type irc[6]://<server>:<port>
					KviStr tmp(KviStr::Format,"irc%s//%s:%u",
							m_pConnectionInfo->pIrcServer->isIpV6() ? "6:" : ":",
							m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),m_pConnectionInfo->pIrcServer->m_uPort);

					g_pApp->addRecentServer(tmp.ptr());

					// save the last server this console used
					if(m_pLastIrcServer)delete m_pLastIrcServer;
					m_pLastIrcServer = new KviIrcServer(*(m_pConnectionInfo->pIrcServer));

					// Ok...we're loggin in now
					setState(LoggingIn);
					resolveLocalHost();
					loginToIrcServer();
				}
				break;
				case KviIrcLink::Ready:
				{
					if(!m_pConnectionInfo)return; // <-- this may happen ONLY if this KviConsole is deleted (called from the destructor)

					if(m_pMonitorList)
					{
						for(KviIrcLinkMonitor *m =m_pMonitorList->first();m;m =m_pMonitorList->next())
						{
							m->connectionTerminated();
						}
					}

					// Prepare data for an eventual reconnect

					KviIrcServer oldServer(*(m_pConnectionInfo->pIrcServer));
					KviStr oldNickname = m_pConnectionInfo->szNickName;

					bool bUnexpectedDisconnect = (!(m_pConnectionInfo->bSentQuit)) && (m_state == Connected);
//					bool bConnectFailureBecauseOfCachedIp = (m_pServerLink->state() == KviIrcLink::Connecting) && m_pConnectionInfo->pIrcServer->cacheIp();

					KviStr szReconnectCmd;

					if(bUnexpectedDisconnect && KVI_OPTION_BOOL(KviOption_boolAutoReconnectOnUnexpectedDisconnect))
					{
						if(KVI_OPTION_BOOL(KviOption_boolRejoinChannelsAfterReconnect))
						{
							for(KviChannel * c = m_pConnectionInfo->pChannelList->first();c;c = m_pConnectionInfo->pChannelList->next())
							{
								// FIXME: password protected channels should have the pass appended
								// FIXME: should join groups of 5 channels per message
								if(szReconnectCmd.isEmpty())szReconnectCmd.append("join ");
								else szReconnectCmd.append(",");
								szReconnectCmd.append(c->name());
							}
						}

						KviStr szQuery;

						if(KVI_OPTION_BOOL(KviOption_boolReopenQueriesAfterReconnect))
						{
							for(KviQuery * q = m_pConnectionInfo->pQueryList->first();q;q = m_pConnectionInfo->pQueryList->next())
							{
								// FIXME: should reopen multiple queries as multiple queries!!!
								// FIXME: should join groups of 5 channels per message
								if(szQuery.isEmpty())szQuery.append("query ");
								else szQuery.append(",");
								szQuery.append(q->target());
							}
						}

						if(szQuery.hasData())
						{
							if(szReconnectCmd.hasData())szReconnectCmd.append(';');
							szReconnectCmd.append(szQuery);
						}

						if(KVI_OPTION_BOOL(KviOption_boolKeepChannelsOpenOnUnexpectedDisconnect))
						{
							keepChannelsOpenAfterDisconnect();
						}

						if(KVI_OPTION_BOOL(KviOption_boolKeepQueriesOpenOnUnexpectedDisconnect))
						{
							keepQueriesOpenAfterDisconnect();
						}
					}

					setState(NotConnected);

					bool bStopOutput = false;
					if(g_pEventManager->hasEventHandlers(KviEvent_OnIrcConnectionTerminated))
					{
						bStopOutput = g_pUserParser->triggerEvent(KviEvent_OnIrcConnectionTerminated,this,0);
					}
					if(!bStopOutput)
					{
						output(KVI_OUT_CONNECTION,__tr2wc("Connection terminated [%s (%s:%u)]"),
								oldServer.hostname().latin1(),
								oldServer.ip().latin1(),
								oldServer.port());
					}

					// do reconnect

					if(bUnexpectedDisconnect && KVI_OPTION_BOOL(KviOption_boolAutoReconnectOnUnexpectedDisconnect))
					{
						m_uConnectAttemptCount = 1;
						output(KVI_OUT_CONNECTION,__tr2wc("The connection terminated unexpectedly. Trying to reconnect..."));
						KviAsyncServerCommandData * d = new KviAsyncServerCommandData;
						d->szServer = oldServer.m_szHostname;
						d->uPort = oldServer.port();
						d->bPortIsOk = true;
						d->bUseIpV6 = oldServer.isIpV6();
						d->bUseSSL = oldServer.useSSL();
						d->szPass = oldServer.password();
						d->szNick = oldNickname;
						d->szInitUMode = oldServer.m_szInitUMode;
						d->szCommandToExecAfterConnect = szReconnectCmd;
						setAsyncServerCommandData(d);
						QTimer::singleShot(1000 * KVI_OPTION_UINT(KviOption_uintAutoReconnectDelay),this,SLOT(asyncServerCommand()));
					}
				}
				break;
				case KviIrcLink::Connecting:
					output(KVI_OUT_CONNECTION,__tr2wc("Contacting %w %s (%s) on port %u"),
						m_pConnectionInfo->pProxy ? __tr2wc("proxy host") : __tr2wc("IRC server"),
						m_pConnectionInfo->pProxy ? m_pConnectionInfo->pProxy->m_szHostname.ptr() : m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
						m_pConnectionInfo->pProxy ? m_pConnectionInfo->pProxy->m_szIp.ptr() : m_pConnectionInfo->pIrcServer->m_szIp.latin1(),
						m_pConnectionInfo->pProxy ? m_pConnectionInfo->pProxy->m_uPort : m_pConnectionInfo->pIrcServer->m_uPort);
				break;
				case KviIrcLink::SSLHandshake:
					output(KVI_OUT_CONNECTION,__tr2wc("Low-level transport connection established [%s (%s:%u)]"),
						m_pConnectionInfo->pProxy ? m_pConnectionInfo->pProxy->m_szHostname.ptr() : m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
						m_pConnectionInfo->pProxy ? m_pConnectionInfo->pProxy->m_szIp.ptr() : m_pConnectionInfo->pIrcServer->m_szIp.latin1(),
						m_pConnectionInfo->pProxy ? m_pConnectionInfo->pProxy->m_uPort : m_pConnectionInfo->pIrcServer->m_uPort);
					outputNoFmt(KVI_OUT_CONNECTION,__tr2wc("Starting Secure Socket Layer handshake"));
				break;
				case KviIrcLink::ProxyLogin:
					output(KVI_OUT_CONNECTION,__tr2wc("%w established [%s (%s:%u)]"),
						m_pLink->usingSSL() ? __tr2wc("Secure proxy connection") : __tr2wc("Proxy connection"),
						m_pConnectionInfo->pProxy->m_szHostname.ptr(),
						m_pConnectionInfo->pProxy->m_szIp.ptr(),
						m_pConnectionInfo->pProxy->m_uPort);
					outputNoFmt(KVI_OUT_CONNECTION,__tr2wc("Negotiating relay information"));
				break;
				case KviIrcLink::ProxyFinalV4:
					outputNoFmt(KVI_OUT_CONNECTION,__tr2wc("Sent connection request, waiting for acknowledgement"));
				break;
				case KviIrcLink::ProxyFinalV5:
					outputNoFmt(KVI_OUT_CONNECTION,__tr2wc("Sent target host data, waiting for acknowledgement"));
				break;
				case KviIrcLink::ProxySelectAuthMethodV5:
					outputNoFmt(KVI_OUT_CONNECTION,__tr2wc("Sent auth method request, waiting for acknowledgement"));
				break;
				case KviIrcLink::ProxyUserPassV5:
					outputNoFmt(KVI_OUT_CONNECTION,__tr2wc("Sent username and password, waiting for acknowledgement"));
				break;
				case KviIrcLink::ProxyFinalHttp:
					outputNoFmt(KVI_OUT_CONNECTION,__tr2wc("Sent connection request, waiting for \"HTTP 200\" acknowledgement"));
				break;
			}
		break;
	}
}

bool KviConsole::isIpV6Connection()
{
	__range_valid(m_pConnectionInfo);
	return m_pConnectionInfo->pIrcServer->isIpV6();
}

void KviConsole::resolveLocalHost()
{
	if(!m_pLink->getLocalHostIp(m_pConnectionInfo->szLocalHostIp,
		m_pConnectionInfo->pIrcServer->isIpV6()))
	{
		bool bGotIp = false;
		if(KVI_OPTION_STRING(KviOption_stringLocalHostIp).hasData())
		{
#ifdef COMPILE_IPV6_SUPPORT
			if(m_pConnectionInfo->pIrcServer->isIpV6())
			{
				if(kvi_isValidStringIp_V6(KVI_OPTION_STRING(KviOption_stringLocalHostIp).ptr()))bGotIp = true;
			} else {
#endif
				if(kvi_isValidStringIp(KVI_OPTION_STRING(KviOption_stringLocalHostIp).ptr()))bGotIp = true;
#ifdef COMPILE_IPV6_SUPPORT
			}
#endif
		}
		if(bGotIp)
		{
			m_pConnectionInfo->szLocalHostIp = KVI_OPTION_STRING(KviOption_stringLocalHostIp);
			output(KVI_OUT_SYSTEMWARNING,__tr2wc("Can't resolve local host address, using user supplied one (%s)"),
				m_pConnectionInfo->szLocalHostIp.ptr());

		} else {
			// FIXME : Maybe check for IPv6 here too ?
			m_pConnectionInfo->szLocalHostIp = "127.0.0.1";
			output(KVI_OUT_SYSTEMWARNING,__tr2wc("Can't resolve local host address, using default 127.0.0.1"),
				m_pConnectionInfo->szLocalHostIp.ptr());
		}
	} else output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Local host address is %s"),m_pConnectionInfo->szLocalHostIp.ptr());
	// For now this is the only we know
	m_pConnectionInfo->szLocalHostFromServer = m_pConnectionInfo->szLocalHostIp;
	m_pConnectionInfo->szLocalIpFromServer = m_pConnectionInfo->szLocalHostIp;
}

void KviConsole::loginToIrcServer()
{
	// Username
	m_pConnectionInfo->pIrcServer->m_szUser.stripWhiteSpace();
	if(m_pConnectionInfo->pIrcServer->m_szUser.isEmpty())
	{
		m_pConnectionInfo->pIrcServer->m_szUser = KVI_OPTION_STRING(KviOption_stringUsername);
		m_pConnectionInfo->pIrcServer->m_szUser.stripWhiteSpace();
		if(m_pConnectionInfo->pIrcServer->m_szUser.isEmpty())
		{
			m_pConnectionInfo->pIrcServer->m_szUser = KVI_DEFAULT_USERNAME;
		}
	} else {
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Using server specific username (%s)"),
			m_pConnectionInfo->pIrcServer->m_szUser.latin1());
	}

	// For now this is the only we know
	m_pConnectionInfo->szUserNameFromServer = m_pConnectionInfo->pIrcServer->m_szUser;
	m_pConnectionInfo->szServerNameFromServer = m_pConnectionInfo->pIrcServer->m_szHostname;

	// Nick stuff
	m_pConnectionInfo->pIrcServer->m_szNick.stripWhiteSpace();
	if(!m_pConnectionInfo->pIrcServer->m_szNick.isEmpty())
	{
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Using server specific nickname (%s)"),
			m_pConnectionInfo->pIrcServer->m_szNick.latin1());
		m_pConnectionInfo->szNickName = m_pConnectionInfo->pIrcServer->m_szNick;
		m_pConnectionInfo->iNicknameUsed = 0;
	} else {
		m_pConnectionInfo->szNickName = KVI_OPTION_STRING(KviOption_stringNickname1);
		m_pConnectionInfo->szNickName.stripWhiteSpace();
		if(m_pConnectionInfo->szNickName.isEmpty())m_pConnectionInfo->szNickName = KVI_DEFAULT_NICKNAME1;
		m_pConnectionInfo->iNicknameUsed = 1;
	}

	// Real name
	m_pConnectionInfo->pIrcServer->m_szRealName.stripWhiteSpace();
	if(!m_pConnectionInfo->pIrcServer->m_szRealName.isEmpty())
	{
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Using server specific real name (%s)"),
			m_pConnectionInfo->pIrcServer->m_szRealName.latin1());
		m_pConnectionInfo->szRealName = m_pConnectionInfo->pIrcServer->m_szRealName;
	} else {
		m_pConnectionInfo->szRealName = KVI_OPTION_STRING(KviOption_stringRealname);
	}

	output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Logging in as %s!%s :%s"),
		m_pConnectionInfo->szNickName.ptr(),
		m_pConnectionInfo->pIrcServer->m_szUser.latin1(),
		m_pConnectionInfo->szRealName.ptr());

	if(!m_pLink->sendFmtData("USER %s 0 %s :%s",
		m_pConnectionInfo->pIrcServer->m_szUser.latin1(),
		m_pConnectionInfo->pIrcServer->m_szHostname.latin1(),
		m_pConnectionInfo->szRealName.ptr()))
	{
		// disconnected in the meantime!
		return;
	}

	// The pass ?
	m_pConnectionInfo->pIrcServer->m_szPass.stripWhiteSpace();
	if(!m_pConnectionInfo->pIrcServer->m_szPass.isEmpty())
	{
		KviStr szHidden;
		int pLen = m_pConnectionInfo->pIrcServer->m_szPass.length();
		for(int i=0;i<pLen;i++)szHidden.append('*');

		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Sending %s as password"),szHidden.ptr());
		if(!m_pLink->sendFmtData("PASS %s",m_pConnectionInfo->pIrcServer->m_szPass.latin1()))
		{
			// disconnected in the meantime
			return;
		}
	}
	if(!m_pLink->sendFmtData("NICK %s",m_pConnectionInfo->szNickName.ptr()))
	{
		// disconnected :(
		return;
	}

	// on connect stuff ?
	KviStr tmp = m_pConnectionInfo->pIrcServer->onConnectCommand();
	tmp.stripWhiteSpace();
	if(tmp.hasData())
	{
		if(KVI_OPTION_BOOL(KviOption_boolBeVerbose))
			output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Executing scheduled server specific \"on connect\" commands"));
		g_pUserParser->parseCommandBuffer(tmp.ptr(),this,0);
	}

	// and wait for the server to agree...
}

void KviConsole::loginComplete(const char * servNick)
{
	if(m_state == Connected)return;
	if(!kvi_strEqualCS(servNick,m_pConnectionInfo->szNickName.ptr()))
	{
		output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("The server refused the suggested nickname (%s) and named you %s instead"),
			m_pConnectionInfo->szNickName.ptr(),servNick);
		m_pConnectionInfo->szNickName = servNick;
	}
	g_pApp->addRecentNickname(servNick);
	setState(Connected);
	bool bHaltOutput = false;
	if(g_pEventManager->hasEventHandlers(KviEvent_OnIrc))
	{
		bHaltOutput = g_pUserParser->triggerEvent(KviEvent_OnIrc,this,0);
	}
	if(!bHaltOutput)outputNoFmt(KVI_OUT_IRC,__tr2wc("Login operations complete, happy ircing!"));

	resurrectDeadQueries();

	// on connect stuff ?
	KviStr tmp = m_pConnectionInfo->pIrcServer->onLoginCommand();
	tmp.stripWhiteSpace();
	if(tmp.hasData())
	{
		if(KVI_OPTION_BOOL(KviOption_boolBeVerbose))
			output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Executing scheduled server specific \"on login\" commands"));
		g_pUserParser->parseCommandBuffer(tmp.ptr(),this,0);
	}

	// Set the configured umode
	KviStr modeStr = m_pConnectionInfo->pIrcServer->initUMode();

	if(modeStr.isEmpty())modeStr = KVI_OPTION_STRING(KviOption_stringDefaultUserMode);

	if(!modeStr.isEmpty())
	{
		if(KVI_OPTION_BOOL(KviOption_boolBeVerbose))
			output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Setting configured user mode"));
		m_pLink->sendFmtData("MODE %s +%s",m_pConnectionInfo->szNickName.ptr(),modeStr.ptr());
	}

	if(KVI_OPTION_BOOL(KviOption_boolUseLagMeterEngine))m_pLagMeter->enable();

	startNotifyList();

	if(KVI_OPTION_BOOL(KviOption_boolShowChannelsJoinOnIrc))
		m_pFrm->executeInternalCommand(KVI_INTERNALCOMMAND_CHANNELSJOIN_OPEN);

	// join saved channels

	if(m_pConnectionInfo->pIrcServer->autoJoinChannelList())
	{
		QStringList * l = m_pConnectionInfo->pIrcServer->autoJoinChannelList();
		for(QStringList::Iterator chanIt = l->begin();chanIt != l->end();++chanIt)
		{
			KviStr tmp = "join ";
			tmp += *chanIt;
			g_pUserParser->parseCommandBuffer(tmp.ptr(),this,0);
		}
	}

	if(m_pConnectionInfo->szCommandToExecAfterConnect.hasData())
	{
		g_pUserParser->parseCommandBuffer(m_pConnectionInfo->szCommandToExecAfterConnect.ptr(),this,0);
	}

	// minimize after connect
	if(KVI_OPTION_BOOL(KviOption_boolMinimizeConsoleAfterConnect))minimize();
}

void KviConsole::closeEvent(QCloseEvent *e)
{
	if(m_state == NotConnected)
	{
		if(m_pFrm->consoleCount() > 1)
		{
			KviWindow::closeEvent(e);
			return;
		} else {
			if(!QMessageBox::warning(this,
					__tr2qs("Confirmation - KVIrc"),
					__tr2qs("You have just attempted to close the last console window.<br>Are you sure you wish to quit KVIrc?"),
					__tr2qs("&Yes"),
					__tr2qs("&No"),
					QString::null, 1, 1))
			{
				g_pApp->quit();
			}
			return;
		}
	}

	if(m_pConnectionInfo)
	{
		if(m_pConnectionInfo->bSentQuit)
		{
			if(m_pFrm->consoleCount() > 1)
			{
				KviWindow::closeEvent(e);
				return;
			}
		}
	}

	e->ignore();
	connectButtonClicked();
}


#define KVI_NUM_NICK_COLORS 95

static const char * g_nickColors[KVI_NUM_NICK_COLORS]=
{
	"0,1"  ,"0,2"  ,"0,3"  ,"0,4"  ,"0,5"  ,"0,6"  ,"0,10" ,"0,12" ,"0,14" , //9
	"1,0"  ,"1,4"  ,"1,7"  ,"1,8"  ,"1,9"  ,"1,11" ,"1,15" ,  //7
	"2,0"  ,"2,4"  ,"2,7"  ,"2,8"  ,"2,9"  ,"2,11" ,"2,15" ,  //7
	"3,8"  ,"3,9"  ,"3,0"  ,"3,15" , //4
	"4,0"  ,"4,1"  ,"4,8"  ,"4,9"  ,"4,11" ,"4,15" , //6
	"5,0"  ,"5,7"  ,"5,8"  ,"5,15" , //4
	"6,0"  ,"6,7"  ,"6,8"  ,"6,9"  ,"6,10" ,"6,11" ,"6,15" , //7
	"7,1"  ,"7,2"  ,"7,5"  ,"7,6"  ,"7,14" , //5
	"8,1"  ,"8,2"  ,"8,3"  ,"8,4"  ,"8,5"  ,"8,6"  ,"8,7"  ,"8,10" ,"8,12" ,"8,14" , //10
	"9,1"  ,"9,2"  ,"9,3"  ,"9,5"  ,"9,6"  ,"9,14" , //6
	"10,1" ,"10,2" , //2
	"11,1" ,"11,2" ,"11,3" ,"11,5" ,"11,6" ,"11,14", //6
	"12,0" ,"12,7" ,"12,8" ,"12,9" ,"12,10","12,11","12,15", //7
	"13,0" ,"13,1" ,"13,6" ,"13,8" ,"13,11","13,15", //6
	"14,0" ,"14,8" ,"14,11","14,15", //4
	"15,1" ,"15,2" ,"15,3" ,"15,6" ,"15,14" //5
};

void KviConsole::outputPrivmsg(KviWindow *wnd,int type,const char *nick,
		const char *user,const char *host,const char *msg,bool bApplyHighlighting)
{
	// FIXME: #warning "THIS IS USED BY WINDOWS THAT ARE NOT BOUND TO THIS IRC CONTEXT"
	// FIXME: #warning "REMEMBER IT IN ESCAPE COMMANDS"
	// __range_valid(wnd);

	if(!nick)nick = currentNickName();

	bool bIsChan = (wnd->type() == KVI_WINDOW_TYPE_CHANNEL);
	bool bIsNotice = ((type == KVI_OUT_CHANNELNOTICE)||(type == KVI_OUT_CHANNELNOTICECRYPTED) \
		   ||(type == KVI_OUT_QUERYNOTICE)||(type == KVI_OUT_QUERYNOTICECRYPTED));

	if(!bApplyHighlighting)goto highlight_done;

	if(KVI_OPTION_BOOL(KviOption_boolAlwaysHighlightNick))
	{
		if(KviStr::ext_contains(msg,currentNickName(),false))
		{
			type = KVI_OUT_HIGHLIGHT;
			if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnHighlight,wnd,nick,user ? user : currentUserName(),host ? host : currentLocalHostName(),msg,currentNickName()))
				return;

			goto highlight_done;
		}
	}

	if(KVI_OPTION_BOOL(KviOption_boolUseWordHighlighting))
	{
		for(QStringList::Iterator it = KVI_OPTION_STRINGLIST(KviOption_stringlistHighlightWords).begin();
				it != KVI_OPTION_STRINGLIST(KviOption_stringlistHighlightWords).end() ; ++it)
		{
			// FIXME : This is SLOOOOOOOOW (QString -> ascii translation!!)
			if(KviStr::ext_contains(msg,(*it).ascii(),false))
			{
				type = KVI_OUT_HIGHLIGHT;
				if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnHighlight,wnd,nick,user ? user : currentUserName(),host ? host : currentLocalHostName(),msg,(*it).ascii()))
					return;

				goto highlight_done;
			}
		}
	}


	if(bIsChan)
	{
		if(((KviChannel *)wnd)->isHighlightedUser(nick))
		{
			type = KVI_OUT_HIGHLIGHT;
			if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnHighlight,wnd,nick,user ? user : currentUserName(),host ? host : currentLocalHostName(),msg,nick))
				return;
			goto highlight_done;
		}
	}

	{
		// FIXME: this is for userhighlighing
		//        maybe mark the users as highlighted in the console user database
		//        and then lookup them there ? this would be potentially a lot faster
		KviRegisteredUser * u = g_pRegisteredUserDataBase->findMatchingUser(nick,user,host);

		if(u)
		{
			if(u->getBoolProperty("highlight"))
			{
				type = KVI_OUT_HIGHLIGHT;
				if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnHighlight,wnd,nick,user ? user : currentUserName(),host ? host : currentLocalHostName(),msg,nick))
					return;
				goto highlight_done;
			}
		}
	}

highlight_done:

	// <PREFIX>nick[!user@host]<POSTFIX>This is a test message

	KviStr szNick;
	szNick.sprintf("\r!n\r%s\r",nick);

	if(KVI_OPTION_BOOL(KviOption_boolShowUserAndHostInPrivmsgView))
	{
		szNick.append(KviStr::Format,"!%s@\r!h\r%s\r",user ? user : currentUserName(),host ? host : currentLocalHostName());
	}

	if(bIsChan && KVI_OPTION_BOOL(KviOption_boolShowChannelUserFlagInPrivmsgView))
	{
		((KviChannel *)wnd)->prependUserFlag(nick,szNick);
	}

	if(KVI_OPTION_BOOL(KviOption_boolColorNicks))
	{
		int sum = 0;
		int i = 0;
		const char * aux = nick;
		while(aux[i])sum += aux[i++];
		szNick.prepend(g_nickColors[sum % KVI_NUM_NICK_COLORS]);
		szNick.prepend(KVI_TEXT_COLOR);
		szNick.append(KVI_TEXT_COLOR);
	}

	if(KVI_OPTION_BOOL(KviOption_boolBoldedNicks))
	{
		szNick.prepend(KVI_TEXT_BOLD);
		szNick.append(KVI_TEXT_BOLD);
	}

	KviWStr szDecodedText(KviWStr::Decode,msg,wnd->textEncoding());

	if(KVI_OPTION_BOOL(KviOption_boolUseExtendedPrivmsgView))
	{
		if(wnd->type() == KVI_WINDOW_TYPE_CHANNEL)
		{
			((KviChannel *)wnd)->outputMessage(type,"%s%s%s%W",
				KVI_OPTION_STRING(KviOption_stringExtendedPrivmsgPrefix).ptr(),
				szNick.ptr(),
				KVI_OPTION_STRING(KviOption_stringExtendedPrivmsgPostfix).ptr(),&szDecodedText);

		} else {
			wnd->output(type,"%s%s%s%W",
				KVI_OPTION_STRING(KviOption_stringExtendedPrivmsgPrefix).ptr(),
				szNick.ptr(),
				KVI_OPTION_STRING(KviOption_stringExtendedPrivmsgPostfix).ptr(),&szDecodedText);
		}
	} else {
		// mini optimisation for speed
		if(bIsChan)
		{
			((KviChannel *)wnd)->outputMessage(type,bIsNotice ? "*%s* %W" : "<%s> %W",szNick.ptr(),&szDecodedText);
		} else {
			wnd->output(type,bIsNotice ? "*%s* %W" : "<%s> %W",szNick.ptr(),&szDecodedText);
		}
	}
}

void KviConsole::avatarChangedUpdateWindows(const char * nick,const char * textLine)
{
	for(KviChannel * c = channelList()->first();c;c = channelList()->next())
	{
		if(c->avatarChanged(nick))
		{
			if(textLine)c->output(KVI_OUT_AVATAR,textLine);
		}
	}
	for(KviQuery * q = queryList()->first();q;q = queryList()->next())
	{
		if(q->avatarChanged(nick))
		{
			if(textLine)q->output(KVI_OUT_AVATAR,textLine);
		}
	}
	m_pNotifyListView->avatarChanged(nick); // recalc the item height here too!
}

void KviConsole::avatarChanged(KviAvatar * avatar,const char * nick,const char * user,const char * host,const char * textLine)
{
	if(!m_pConnectionInfo)return; //ops...

	
	bool bRegisteredStuff = false;

	if(KVI_OPTION_BOOL(KviOption_boolSetLastAvatarAsDefaultForRegisteredUsers))
	{
		// Don't even try to do it for myself
		if(!kvi_strEqualCI(nick,currentNickName()))
		{
			KviRegisteredUser * u = g_pRegisteredUserDataBase->findMatchingUser(nick,user,host);
			if(u)
			{
				if(avatar)u->setProperty("avatar",avatar->localPath());
				else u->setProperty("avatar",0);
				bRegisteredStuff = true;
			}
		}
	}
	
	if(!bRegisteredStuff)
	{
		// cache it
		if(avatar)
			KviAvatarCache::instance()->replace(avatar->localPath(),KviIrcMask(nick,user,host),currentNetworkName());
		else
			KviAvatarCache::instance()->remove(KviIrcMask(nick,user,host),currentNetworkName());
	}
	
	avatarChangedUpdateWindows(nick,textLine);
}

void KviConsole::checkDefaultAvatar(KviIrcUserEntry *e,const char * nick,const char * user,const char * host)
{	
//	KviStr mask(KviStr::Format,"%s!%s@%s",nick,user,host);
	KviRegisteredUser * u = g_pRegisteredUserDataBase->findMatchingUser(nick,user,host);
	if(u)
	{
		// the user is registered...
		KviStr szAvatar;
		if(u->getProperty("avatar",szAvatar))
		{
			// the user has a default avatar...
			KviAvatar * avatar = g_pIconManager->getAvatar(szAvatar.ptr());
			if(avatar)
			{
				e->setAvatar(avatar);
				avatarChangedUpdateWindows(nick,0);
				return;
			}
		}
	}
	
	// look it up in the cache then
	QString szCached = KviAvatarCache::instance()->lookup(KviIrcMask(nick,user,host),currentNetworkName());
	if(!szCached.isEmpty())
	{
		// got a cache hit... is it on disk ?
		KviAvatar * avatar = g_pIconManager->getAvatar(szCached.latin1());
		if(avatar)
		{
			// cached image on disk
			e->setAvatar(avatar);
			avatarChangedUpdateWindows(nick,0);
			return;
		} else {
			// no cached image on disk.. will need to requery it anyway
			// remove from cache
			KviAvatarCache::instance()->remove(KviIrcMask(nick,user,host),currentNetworkName());
		}
	}
}

void KviConsole::resetAvatarForMatchingUsers(KviRegisteredUser * u)
{
	if(!m_pConnectionInfo)return;

	KviStr szPath;
	u->getProperty("avatar",szPath);

	QAsciiDictIterator<KviIrcUserEntry> it(*(m_pConnectionInfo->pUserDataBase->dict()));
	while(KviIrcUserEntry * e = it.current())
	{
		if(e->hasHost())
		{
			if(u->matchesFixed(it.currentKey(),e->user(),e->host()))
			{
				KviAvatar * a = 0;
				if(szPath.hasData())a = g_pIconManager->getAvatar(szPath.ptr());
				e->setAvatar(a);
				avatarChangedUpdateWindows(it.currentKey(),0);
			}
		}
		++it;
	}
}

KviAvatar * KviConsole::setAvatar(const char * nick,const char * user,const char * host,const char * fName)
{
	if(!m_pConnectionInfo)return 0;
	KviIrcUserEntry * e = m_pConnectionInfo->pUserDataBase->find(nick);
	if(e)
	{
		// User and host must match
		if(user && e->hasUser())
		{
			if(!kvi_strEqualCI(user,e->user()))return 0;
		}

		if(host && e->hasHost())
		{
			if(!kvi_strEqualCI(host,e->host()))return 0;
		}

		// Ok...got it

		KviAvatar * avatar = g_pIconManager->getAvatar(fName);
		if(avatar)
		{
			e->setAvatar(avatar);
			avatarChanged(avatar,nick,user,host,0);
			return avatar;
		}
	}
	return 0;
}

KviAvatar * KviConsole::defaultAvatarFromOptions()
{
	QPixmap * avatar = KVI_OPTION_PIXMAP(KviOption_pixmapMyAvatar).pixmap();
	if(!avatar)return 0;
	if(avatar->isNull())return 0;
	if(KVI_OPTION_STRING(KviOption_stringMyAvatar).isEmpty())return 0;
	return new KviAvatar(new QPixmap(*avatar),KVI_OPTION_PIXMAP(KviOption_pixmapMyAvatar).path(),KVI_OPTION_STRING(KviOption_stringMyAvatar).ptr());
}

KviAvatar * KviConsole::currentAvatar()
{
	// FIXME: this doesn't work if I'm not on any channel and have no query open....
	//        a caching user_data_base ?
	if(!m_pConnectionInfo)return 0;
	KviIrcUserEntry * e = m_pConnectionInfo->pUserDataBase->find(currentNickName());
	if(!e)return 0;
	KviAvatar * a = e->avatar();
	if(!a)
	{
		a = defaultAvatarFromOptions();
		if(a)
		{
			e->setAvatar(a);
			avatarChanged(a,currentNickName(),0,0,0);
		}
	}
	return a;
}

void KviConsole::applyOptions()
{
	m_pNotifyListView->applyOptions();
	m_pInput->applyOptions();
	m_pIrcView->applyOptions();

	KviWindow::applyOptions();

	// trick
	resize(width() - 1,height() - 1);
	resize(width() + 1,height() + 1);
}

#include "kvi_console.moc"
