//   File : libkvilag.cpp
//   Creation date :  Oct 10 01:06:09 CEST 2002 by Juanjo lvarez
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Juanjo lvarez (juanjux at yahoo.es)
//   Copyright (C) 2002 Szymon Stefanek (kvirc at tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//


#include "kvi_module.h"
#include "kvi_uparser.h"
#include "kvi_console.h"
#include "kvi_locale.h"


//-------------------------------------------------
/*
    @doc: lag.enable
    @type:
        command
    @title:
        lag.enable
    @short:
        Enable the lag metering engine
    @syntax:
        lag.enable
    @description:
        This command enable the lag metering engine, enabling this way the other commands of this[br]
        module.[br]
    @seealso:
        [cmd]lag.disable[/cmd]
        [cmd]lag.lag[/cmd]
        [cmd]lag.setinterval[/cmd]
        [cmd]lag.getinterval[/cmd]
        [cmd]lag.getalarmtime[/cmd]
        [cmd]lag.setalarmtime[/cmd]
        [cmd]lag.removelaget[/cmd]
*/
//-------------------------------------------
// lag.enable
//-------------------------------------------
static bool lag_module_cmd_enable(KviModule *m,KviCommand *c)
{
    ENTER_CONTEXT(c,"lag_module_cmd_enable");
    if( (!c->window()->console()) || c->window()->console()->isNotConnected() )return c->noIrcContext();
    if(c->window()->console()->lagMeter()->isEnabled())
    {
        c->warning(__tr("Lag meter already enabled"));
        return false;
    }

    c->window()->console()->lagMeter()->enable();
    return c->leaveContext();
}
//-------------------------------------------------
/*
    @doc: lag.disable
    @type:
        command
    @title:
        lag.disable
    @short:
        Disable the lag metering engine
    @syntax:
        lag.disable
    @description:
        This command disable the lag metering engine, thus disabling the other commands of this[br]
        module (except lag.enable).[br]
    @seealso:
        [cmd]lag.enable[/cmd]
        [cmd]lag.lag[/cmd]
        [cmd]lag.setinterval[/cmd]
        [cmd]lag.getinterval[/cmd]
        [cmd]lag.getalarmtime[/cmd]
        [cmd]lag.setalarmtime[/cmd]
        [cmd]lag.removelaget[/cmd]
*/
//-------------------------------------------
// lag.disable
//-------------------------------------------

static bool lag_module_cmd_disable(KviModule *m,KviCommand *c)
{
    ENTER_CONTEXT(c,"lag_module_cmd_disable");
    if(!c->window()->console()->lagMeter()->isEnabled())
    {
        c->warning(__tr("Lag meter was not enabled"));
        return false;
    }

    c->window()->console()->lagMeter()->disable();
    return c->leaveContext();
}
//-------------------------------------------------
/*
    @doc: $lag.lag
    @type:
        function
    @title:
        $lag.lag
    @short:
        Returns the lag on the current server
    @syntax:
        $lag.lag
    @description:
        This function returns the lag in the current server, in miliseconds.[br]
    @seealso:
        [cmd]lag.enable[/cmd]
        [cmd]lag.disable[/cmd]
        [cmd]lag.setinterval[/cmd]
        [cmd]lag.getinterval[/cmd]
        [cmd]lag.getalarmtime[/cmd]
        [cmd]lag.setalarmtime[/cmd]
        [cmd]lag.removelaget[/cmd]
*/
//-------------------------------------------
// lag.lag
//-------------------------------------------

static bool lag_module_fnc_lag(KviModule *m,KviCommand *c,KviParameterList *parms,KviStr &buffer)
{
    ENTER_CONTEXT(c,"lag_module_fnc_lag");
    if( (!c->window()->console()) || c->window()->console()->isNotConnected() )return c->noIrcContext();

    if(!c->window()->console()->lagMeter()->isEnabled())
    {
        c->warning(__tr("Lag meter was not enabled"));
        return false;
    }

    buffer.append(KviStr::Format,"%d", c->window()->console()->lagMeter()->getLag());

    return c->leaveContext();
}
//-------------------------------------------------
/*
    @doc: lag.getinterval
    @type:
        function
    @title:
        $lag.getinterval
    @short:
        Returns the lag checking interval
    @syntax:
        $lag.getinterval
    @description:
        This function return the interval, in miliseconds, between lag checks by the lag metering engine.[br]
    @seealso:
        [cmd]lag.enable[/cmd]
        [cmd]lag.disable[/cmd]
        [cmd]lag.lag[/cmd]
        [cmd]lag.setinterval[/cmd]
        [cmd]lag.getalarmtime[/cmd]
        [cmd]lag.setalarmtime[/cmd]
        [cmd]lag.removelaget[/cmd]
*/
//-------------------------------------------
// lag.getinterval
//-------------------------------------------
static bool lag_module_fnc_getinterval(KviModule *m,KviCommand *c,KviParameterList *parms,KviStr &buffer)
{
    ENTER_CONTEXT(c,"lag_module_fnc_getinterval");
    //It doesn't need the console to be connected
    if( (!c->window()->console()) )return c->noIrcContext();

    buffer.append(KviStr::Format,"%d",c->window()->console()->lagMeter()->getLagCheckInterval());
    return c->leaveContext();
}
//-------------------------------------------------
/*
    @doc: lag.setinterval
    @type:
        command
    @title:
        lag.setinterval
    @short:
        Sets the lag checking interval
    @syntax:
        lag.setinterval <time>
    @description:
        This command sets the interval, in miliseconds, between lag checks by the lag metering engine.[br]
        This command will change that interval only for the current console. If you want to permanently[br]
        change the interval for all consoles you will have to set the option KviOption_UIntLagCheckingInterval.[br]
    @seealso:
        [cmd]lag.enable[/cmd]
        [cmd]lag.disable[/cmd]
        [cmd]lag.lag[/cmd]
        [cmd]lag.getinterval[/cmd]
        [cmd]lag.getalarmtime[/cmd]
        [cmd]lag.setalarmtime[/cmd]
        [cmd]lag.removelaget[/cmd]
*/
//-------------------------------------------
// lag.setinterval
//-------------------------------------------
static bool lag_module_cmd_setinterval(KviModule *m,KviCommand *c)
{
    ENTER_CONTEXT(c,"lag_module_cmd_enable");
    //It doesn't need the console to be connected
    if( (!c->window()->console()) )return c->noIrcContext();
    KviStr szInterval;
    if(!g_pUserParser->parseCmdFinalPart(c,szInterval))return false;
    int interval = szInterval.toInt();
    c->window()->console()->lagMeter()->setLagCheckInterval(interval);
    return c->leaveContext();
}
//-------------------------------------------------
/*
    @doc: lag.getalarmtime
    @type:
        function
    @title:
        $lag.getalarmtime
    @short:
        Returns the lag threshold that triggers the lag alarm
    @syntax:
        $lag.getalarmtime
    @description:
        This function returns the lag threshold, in miliseconds that triggers the lag alarm (in the form[br]
        of the OnLagAlarmTimeUp event).[br]
    @seealso:
        [cmd]lag.enable[/cmd]
        [cmd]lag.disable[/cmd]
        [cmd]lag.lag[/cmd]
        [cmd]lag.setinterval[/cmd]
        [cmd]lag.getinterval[/cmd]
        [cmd]lag.setalarmtime[/cmd]
        [cmd]lag.removelaget[/cmd]
*/
//-------------------------------------------
// lag.getalarmtime
//-------------------------------------------
static bool lag_module_fnc_getalarmtime(KviModule *m,KviCommand *c,KviParameterList *parms,KviStr &buffer)
{
    ENTER_CONTEXT(c,"lag_module_fnc_getalarmtime");
    //It doesn't need the console to be connected
    if( (!c->window()->console()) )return c->noIrcContext();

    buffer.append(KviStr::Format,"%d",c->window()->console()->lagMeter()->getLagAlarmTime());
    return c->leaveContext();
}
//-------------------------------------------------
/*
    @doc: lag.setalarmtime
    @type:
        command
    @title:
        lag.setalarmtime
    @short:
        Sets the lag threshold that triggers the lag alarm.
    @syntax:
        lag.setalarmtime <time>
    @description:
        This command sets the lag threshold, in miliseconds, that triggers the lag alarm (in the form[br]
        of the OnLagAlarmTimeUp event).[br]
        This command will change that time only for the current console. If you want to permanently[br]
        change the interval for all console you will have to set the option KviOption_UIntLagAlarmTime.[br]
    @seealso:
        [cmd]lag.enable[/cmd]
        [cmd]lag.disable[/cmd]
        [cmd]lag.lag[/cmd]
        [cmd]lag.getinterval[/cmd]
        [cmd]lag.setinterval[/cmd]
        [cmd]lag.getalarmtime[/cmd]
        [cmd]lag.removelaget[/cmd]
*/
//-------------------------------------------
// lag.setalarmtime
//-------------------------------------------
static bool lag_module_cmd_setalarmtime(KviModule *m,KviCommand *c)
{
    ENTER_CONTEXT(c,"lag_module_cmd_setalarmtime");
    //It doesn't need the console to be connected
    if( (!c->window()->console()) )return c->noIrcContext();

    KviStr szAlarm;
    if(!g_pUserParser->parseCmdFinalPart(c,szAlarm))return false;
    int alarm = szAlarm.toInt();
    c->window()->console()->lagMeter()->setLagAlarmTime(alarm);
    return c->leaveContext();
}

//------------------------------------------------
static bool lag_module_init(KviModule * m)
{
	m->registerCommand("enable", lag_module_cmd_enable);
    m->registerCommand("disable", lag_module_cmd_disable);
    m->registerFunction("lag", lag_module_fnc_lag);
    m->registerFunction("getinterval", lag_module_fnc_getinterval);
    m->registerCommand("setinterval", lag_module_cmd_setinterval);
    m->registerFunction("getalarmtime", lag_module_fnc_getalarmtime);
    m->registerCommand("setalarmtime", lag_module_cmd_setalarmtime);

	return true;
}
//-------------------------------------------------
static bool lag_module_cleanup(KviModule *m)
{
	return true;
}
//-------------------------------------------------
static bool lag_module_can_unload(KviModule *m)
{
    return true;
}
//-------------------------------------------------
KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"Lag",                                                 // module name
	"1.0.0",                                                // module version
	"          (C) 2002 Juanjo Alvarez (juanjux@yahoo.es)", // author & (C)
	"Functions and commands to control the lag metering engine",
	lag_module_init,
	0,
	0,
	lag_module_cleanup,
};
