//
//   File : optw_nickserv.cpp
//   Creation date : Fri Aug 10 2001 03:38:10 CEST by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2001 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "optw_nickserv.h"

#include "kvi_settings.h"

#include <qlayout.h>
#include <qtooltip.h>
#include <qasciidict.h>

#include "kvi_options.h"
#include "kvi_locale.h"
#include "kvi_nickserv.h"
#include "kvi_ircmask.h"

// kvi_app.cpp
extern KVIRC_API KviNickServDataBase * g_pNickServDataBase;


KviNickServOptionsWidget::KviNickServOptionsWidget(QWidget * parent)
: KviOptionsWidget(parent,"nickserv_options_widget")
{
	m_pCurrentItem = 0;

	m_pContextPopup = new QPopupMenu(this);

	createLayout(3,1);

	m_pEnableSelector = addBoolSelector(0,0,1,0,__tr2qs("Enable automatic NickServ identification"),KviOption_boolEnableAutomaticNickServIdentification);
	connect(m_pEnableSelector,SIGNAL(toggled(bool)),this,SLOT(enableToggled(bool)));

	m_pListView = new QListView(this);
	m_pListView->addColumn(__tr2qs("Nickname"));
	m_pListView->addColumn(__tr2qs("Server"));
	m_pListView->addColumn(__tr2qs("NickServ Mask"));
	m_pListView->addColumn(__tr2qs("Identify Command"));
	m_pListView->setAllColumnsShowFocus(true);
	m_pListView->setSelectionMode(QListView::Single);

	connect(m_pListView,SIGNAL(mouseButtonPressed(int,QListViewItem *,const QPoint &,int)),this,SLOT(itemClicked(int,QListViewItem *,const QPoint &,int)));
	connect(m_pListView,SIGNAL(selectionChanged(QListViewItem *)),this,SLOT(selectionChanged(QListViewItem *)));
	addWidgetToLayout(m_pListView,0,1,0,1);

#ifdef COMPILE_INFO_TIPS
	QToolTip::add(m_pListView,
		__tr2qs("<center>This is a list of NickServ identification rules. " \
			"KVIrc will use them to model its automatic interaction with NickServ.<br>" \
			"Please be aware that this feature can cause your NickServ passwords to be stolen " \
			"if used improperly. Follow the suggestions in the tooltips and " \
			"make sure that you fully understand the NickServ authentication protocol.<br>" \
			"In other words, be sure to know what you're doing.<br>" \
			"Also note that the password that you provide is stored as <b>PLAIN TEXT</b>." \
			"Right-click to add/remove rules.</center>"));
#endif

	m_pListView->setEnabled(KVI_OPTION_BOOL(KviOption_boolEnableAutomaticNickServIdentification));

	m_pRuleBox = addGroupBox(0,2,0,2,2,QGroupBox::Horizontal,__tr2qs("Identification Rule"));

	m_pNicknameLabel = new QLabel(__tr2qs("Nickname:"),m_pRuleBox);
	m_pNicknameEdit = new QLineEdit(m_pRuleBox);
#ifdef COMPILE_INFO_TIPS
	QToolTip::add(m_pNicknameEdit,
		__tr2qs("<center>This is the nickname that this rule will attempt to authenticate with NickServ, " \
			"i.e. the nickname that you have registered with NickServ</center>"));
#endif

	m_pServerMaskLabel = new QLabel(__tr2qs("Server:"),m_pRuleBox);
	m_pServerMaskEdit = new QLineEdit(m_pRuleBox);
#ifdef COMPILE_INFO_TIPS
	QToolTip::add(m_pServerMaskEdit,
		__tr2qs("<center>This is a mask that defines the server(s) that this identification rule will apply to. " \
			"For a single server, this option should just be a hostname like <b>irc.webchat.org</b>. " \
			"You can use wildcards to match multiple servers, e.g. <b>*.freenode.net</b> will work on the whole freenode network " \
			"(by matching any server names that end in \".freenode.net\").<br>" \
			"Please note that some servers have more than one name. For example, <b>irc.eu.dal.net</b> redirects to a \"random\" " \
			"DALNet server, so the real server name will be different. You must provide the name that " \
			"the server calls itself, usually displayed in the console window title.</center>"));
#endif


	m_pNickServMaskLabel = new QLabel(__tr2qs("NickServ mask:"),m_pRuleBox);
	m_pNickServMaskEdit = new QLineEdit(m_pRuleBox);
#ifdef COMPILE_INFO_TIPS
	QToolTip::add(m_pNickServMaskEdit,
		__tr2qs("<center>This is the mask that NickServ must match to be correctly identified as the NickServ service. "  \
			"This usually will be something like <b>NickServ!service@services.dalnet</b>.<br>" \
			"You can use wildcards for this field, but generally it is a security flaw. " \
			"If you're 100% sure that NO user on the network can use the nickname \"NickServ\", " \
			"the mask <b>NickServ!*@*</b> may be safe to use in this field.</center>"));
#endif

	m_pIdentifyCommandLabel = new QLabel(__tr2qs("Identify command:"),m_pRuleBox);
	m_pIdentifyCommandEdit = new QLineEdit(m_pRuleBox);
#ifdef COMPILE_INFO_TIPS
	QToolTip::add(m_pIdentifyCommandEdit,
		__tr2qs("<center>This is the command that will be executed when NickServ requests authentication " \
			"for the nickname described in this rule (if the both server and NickServ mask are matched). " \
			"This usually will be something like <b>msg NickServ identify &lt;yourpassword&gt;</b>.<br>" \
			"You can use <b>msg -q</b> if you don't want the password echoed on the screen. " \
			"Please note that there is no leading slash in this command.</center>"));
#endif

	enableToggled(m_pEnableSelector->isChecked());
	//layout()->setRowStretch(3,1);


	QListViewItem * item = 0;

	QAsciiDictIterator<KviNickServEntryList> it(*g_pNickServDataBase->entryDict());
	while(KviNickServEntryList * l = it.current())
	{
		for(KviNickServEntry * e = l->first();e;e = l->next())
		{
			item = new QListViewItem(m_pListView,it.currentKey(),e->szServerMask.ptr(),e->szNickServMask.ptr(),e->szIdentifyCommand.ptr());
		}
		++it;
	}
}

KviNickServOptionsWidget::~KviNickServOptionsWidget()
{
}

void KviNickServOptionsWidget::saveLastItem()
{
	if(!m_pCurrentItem)return;
	KviStr tmp = m_pNicknameEdit->text();
	if(tmp.isEmpty())tmp = "???";
	m_pCurrentItem->setText(0,tmp.ptr());

	tmp = m_pServerMaskEdit->text();
	if(tmp.isEmpty())tmp = "???";
	m_pCurrentItem->setText(1,tmp.ptr());

	tmp = m_pNickServMaskEdit->text();
	KviIrcMask mask(tmp.ptr());
	tmp.sprintf("%s!%s@%s",mask.nick(),mask.user(),mask.host());
	m_pCurrentItem->setText(2,tmp.ptr());

	m_pCurrentItem->setText(3,m_pIdentifyCommandEdit->text());
}

void KviNickServOptionsWidget::selectionChanged(QListViewItem * it)
{
	saveLastItem();
	m_pCurrentItem = it;
	enableToggled(m_pEnableSelector->isChecked());
	if(it)
	{
		if(m_pEnableSelector->isChecked())
			m_pRuleBox->setEnabled(true);
		m_pNicknameEdit->setText(it->text(0));
		m_pServerMaskEdit->setText(it->text(1));
		m_pNickServMaskEdit->setText(it->text(2));
		m_pIdentifyCommandEdit->setText(it->text(3));
	} else {
		m_pRuleBox->setEnabled(false);
		m_pNicknameEdit->setText("");
		m_pServerMaskEdit->setText("");
		m_pNickServMaskEdit->setText("");
		m_pIdentifyCommandEdit->setText("");
	}
}

void KviNickServOptionsWidget::itemClicked(int btn,QListViewItem * it,const QPoint &pnt,int)
{
	if(btn & RightButton)
	{
		m_pContextPopup->clear();
		int id = m_pContextPopup->insertItem(__tr2qs("&New Rule"),this,SLOT(newRule()));
		id = m_pContextPopup->insertItem(__tr2qs("Re&move Rule"),this,SLOT(deleteRule()));
		m_pContextPopup->setItemEnabled(id,it);

		// context
		m_pContextPopup->popup(pnt);
	}
}

void KviNickServOptionsWidget::enableToggled(bool b)
{
	m_pListView->setEnabled(b);
	m_pRuleBox->setEnabled(b && m_pCurrentItem);
}

void KviNickServOptionsWidget::commit()
{
	saveLastItem();
	g_pNickServDataBase->clear();

	for(QListViewItem * it = m_pListView->firstChild();it;it = it->nextSibling())
	{
		KviStr szNick = it->text(0);
		KviStr szServ = it->text(1);
		KviStr szNSer = it->text(2);
		KviStr szICmd = it->text(3);
		g_pNickServDataBase->insertEntry(szNick.ptr(),szServ.ptr(),szNSer.ptr(),szICmd.ptr());
	}
	KviOptionsWidget::commit();
}

void KviNickServOptionsWidget::newRule()
{
	saveLastItem();
	QListViewItem * it = new QListViewItem(	m_pListView,
		KVI_OPTION_STRING(KviOption_stringNickname1).ptr(),
		"irc.unknown.net",
		"NickServ!s@services.unknown.net",
		"msg NickServ IDENTIFY yourpassword");
	m_pListView->setCurrentItem(it);
	m_pListView->setSelected(it,true);
}

void KviNickServOptionsWidget::deleteRule()
{
	if(m_pCurrentItem)
	{
		QListViewItem * it = m_pCurrentItem;
		m_pCurrentItem = 0;
		delete it;
	}
	if(!m_pCurrentItem)
	{
		QListViewItem * f = m_pListView->firstChild();
		if(f)
		{
			m_pListView->setSelected(f,true);
			m_pListView->setCurrentItem(f);
		} else selectionChanged(0);
	}
}

#include "m_optw_nickserv.moc"
