//
//   File : libkvitoolbar.cpp
//   Creation date : Wed 6 Nov 22:51:59 2002 GMT by Szymon Stefanek
//
//   This toolbar is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "kvi_module.h"
#include "kvi_uparser.h"
#include "kvi_locale.h"
#include "kvi_command.h"
#include "kvi_window.h"
#include "kvi_app.h"
#include "kvi_error.h"
#include "kvi_cmdformatter.h"

#ifdef COMPILE_SCRIPTTOOLBAR
	#include "kvi_scripttoolbar_manager.h"
	#include "kvi_scripttoolbar_definition.h"
	#include "kvi_scripttoolbar_itemdefinition.h"
	#include "kvi_scripttoolbar.h"
#endif // COMPILE_SCRIPTTOOLBAR

/*
	@doc: toolbar.create
	@type:
		command
	@title:
		toolbar.create
	@keyterms:
		Creating toolbars
	@short:
		Creates a toolbar
	@syntax:
		toolbar.create <name> <label>
	@description:
		Creates a toolbar named <name> and with the visible label (caption) <label>.[br]
		If a toolbar with the same name already exists then this command is equivalent to [cmd]toolbar.setlabel[/cmd].[br]
		This command doesn't show the toolbar in any of the KVIrc frames (you need to use [cmd]toolbar.show[/cmd]).[br]
		Toolbars are saved and preserved across KVIrc sessions thus you don't need to create them
		on application startup: you only create them once.
		The toolbar will exist until [cmd]toolbar.destroy[/cmd] is called (you might want to add
		an "uninstall" feature to your script).[br]
		If the feature "scripttoolbar" wasn't enabled at compile time then this command does nothing.
	@seealso:
		[fnc]$toolbar.exists[/fnc]()
		[cmd]toolbar.remove[/cmd]
		[cmd]toolbar.show[/cmd]
		[cmd]toolbar.destroy[/cmd]
*/

static bool toolbar_module_cmd_create(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::create");
	KviStr szName;
	KviStr szLabel;
	if(!g_pUserParser->parseCmdSingleToken(c,szName))return false;
	if(!g_pUserParser->parseCmdFinalPart(c,szLabel))return false;
#ifdef COMPILE_SCRIPTTOOLBAR

	if(szName.isEmpty())
	{
		c->warning(__tr("No toolbar name specified"));
		return c->leaveContext();
	}

	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->get(QString(szName.ptr()));

	def->setLabel(QString(szLabel.ptr()));

#endif // !COMPILE_SCRIPTTOOLBAR
	return c->leaveContext();
}



/*
	@doc: toolbar.define
	@type:
		command
	@title:
		toolbar.define
	@keyterms:
		Creating toolbars (quickly!)
	@short:
		Defines a toolbar (creates it)
	@syntax:
		toolbar.define [-m] (<tbname>,<label>)
		{
			button [-d] (<name>,<icon>,<text>)<callback code>
			separator <sepname>
			...
			...
			...
		}
	@description:
		This command provides a quick way to define a toolbar: it is meant as a shortcut for
		the sequence [cmd]toolbar.create[/cmd],[cmd]toolbar.addbutton[/cmd],....[br]
		<tbname> and <label> have the same meaning as for the [cmd]toolbar.create[/cmd] command.[br]
		The definition block can contain one or more <b>button</b> statements that
		act just like [cmd]toolbar.addbutton[/cmd] was called. The <name>,<icon>,<text> and <callback code>
		parameters have all the same meaning.[br]
		The definition block can contains one or more <b>separator</b> statements that
		act exactly like [cmd]toolbar.addseparator[/cmd].
		If the toolbar identified by <tbname> already exists it is cleared, unless -m is used: in that
		case the new contents are merged to the existing ones.[br]
	@seealso:
		[cmd]toolbar.create[/cmd]
		[cmd]toolbar.addbutton[/cmd]
		[cmd]toolbar.addseparator[/cmd]
*/

static bool toolbar_module_cmd_define(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::define");


	KviParameterList paramList;
	paramList.setAutoDelete(true);

	c->skipWhiteSpace();
	if(*(c->ptr()) != '(')return c->error(KviError_openParenthesisExpected);

	if(!g_pUserParser->extractFunctionParameters(c,&paramList))return false;

	c->skipWhiteSpace();

	if(*(c->ptr()) != '{')return c->error(KviError_openBraceExpected);


#ifdef COMPILE_SCRIPTTOOLBAR

	KviStr toolbar = paramList.safeFirstParam();
	KviStr label   = paramList.safeNextParam();

	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->get(QString(toolbar.ptr()));
	def->setLabel(QString(label.ptr()));

	if(!c->hasSwitch('m'))def->clear();

#endif //COMPILE_SCRIPTTOOLBAR

	c->next();
	c->skipWhiteSpace();

	while(*(c->ptr()) && (*(c->ptr()) != '}'))
	{
		if(*(c->ptr()) == '#')g_pUserParser->skipComment(c);
		else {
			if(kvi_strEqualCIN(c->ptr(),"button",6))
			{
				c->skip(6);
				c->skipWhiteSpace();

				bool bDisabled = false;

				while(*(c->ptr()) == '-')
				{
					c->next();
					if(*(c->ptr()) == 'd')bDisabled = true;
					c->next();
					c->skipWhiteSpace();
				}

				if(*(c->ptr()) != '(')return c->error(KviError_openParenthesisExpected);

				paramList.clear();

				if(!g_pUserParser->extractFunctionParameters(c,&paramList))return false;

				c->skipWhiteSpace();

				//if(*(c->ptr()) != '{')return c->error(KviError_openBraceExpected);

				char * aux = c->ptr();
				if(!g_pUserParser->skipCommand(c))return false;

				KviStr body(aux,c->ptr());
				KviCommandFormatter::bufferFromBlock(body);

#ifdef COMPILE_SCRIPTTOOLBAR
				KviStr btname = paramList.safeFirstParam();
				KviStr bticon = paramList.safeNextParam();
				KviStr bttext = paramList.safeNextParam();

				if(btname.isEmpty())
				{
					c->warning(__tr("No button name specified!"));
				} else {
					KviScriptToolBarItemDefinition * i = def->getItemDefinition(QString(btname.ptr()),KviScriptToolBarItemDefinition::Button);
		
					if(i->type() != KviScriptToolBarItemDefinition::Button)
					{
						c->warning(__tr("The specified name is already assigned to a different type item (%s)"),btname.ptr());
					} else {
						if(bticon.hasData())i->setIcon(QString(bticon.ptr()));
						if(bttext.hasData())i->setText(QString(bttext.ptr()));
						i->setCode(QString(body.ptr()));
						i->setEnabled(!bDisabled);
					}
				}
#endif //COMPILE_SCRIPTTOOLBAR

			} else if(kvi_strEqualCIN(c->ptr(),"separator",9))
			{
				c->skip(9);
				c->skipSpace();
				KviStr sname;
				if(!g_pUserParser->parseCmdFinalPart(c,sname))return false;
#ifdef COMPILE_SCRIPTTOOLBAR
				if(sname.isEmpty())
				{
					c->warning(__tr("No separator name specified"));
				} else {
					KviScriptToolBarItemDefinition * i = def->getItemDefinition(QString(sname.ptr()),KviScriptToolBarItemDefinition::Separator);

					if(i->type() != KviScriptToolBarItemDefinition::Separator)
					{
						c->warning(__tr("The specified name is assigned to an item that is not a separator (%s)"),sname.ptr());
					}
				}
#endif //COMPILE_SCRIPTTOOLBAR
			} else return c->error(KviError_unexpectedToken);
		}
		c->skipWhiteSpace();
	}

	if(!(*(c->ptr())))return c->error(KviError_missingClosingBrace);

	c->next();

	return c->leaveContext();
}



/*
	@doc: toolbar.destroy
	@type:
		command
	@title:
		toolbar.destroy
	@keyterms:
		Destroying a toolbar
	@short:
		Destroys a toolbar
	@syntax:
		toolbar.destroy <name>
	@description:
		Destroys the toolbar named <name>.[br]
		The toolbar definition is definitively destroyed with all the items contained.[br]
		If the feature "scripttoolbar" wasn't enabled at compile time then this command does nothing.
	@seealso:
		[fnc]$toolbar.exists[/fnc]()
		[cmd]toolbar.create[/cmd]
*/

static bool toolbar_module_cmd_destroy(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::destroy");
	KviStr szName;
	if(!g_pUserParser->parseCmdFinalPart(c,szName))return false;
#ifdef COMPILE_SCRIPTTOOLBAR
	g_pScriptToolBarManager->remove(QString(szName.ptr()));
#endif // !COMPILE_SCRIPTTOOLBAR
	return c->leaveContext();
}


/*
	@doc: toolbar.show
	@type:
		command
	@title:
		toolbar.show
	@short:
		Shows a toolbar
	@syntax:
		toolbar.show [-q] [-a] <name>
	@description:
		Shows the toolbar named <name>.[br]
		If -a is specified then the toolbar is shown in all the frames, otheriwse it is
		shown in the current frame only.[br]
		If -q is specified the command runs in quiet mode and does not complain if the
		specified toolbar doesn't exist.[br]
		Please note that you don't need to show the toolbar at application startup: KVIrc will
		take care of this for you (if the user chooses to keep the toolbar visible).[br]
		You should execute this command only if you create the toolbar after KVIrc has been
		started.[br]
		If the feature "scripttoolbar" wasn't enabled at compile time then this command does nothing.
	@seealso:
		[fnc]$toolbar.exists[/fnc]()
		[cmd]toolbar.create[/cmd]
*/

static bool toolbar_module_cmd_show(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::show");
	KviStr szName;
	if(!g_pUserParser->parseCmdFinalPart(c,szName))return false;

#ifdef COMPILE_SCRIPTTOOLBAR
	if(szName.isEmpty())
	{
		if(!c->hasSwitch('q'))c->warning(__tr("No toolbar name specified"));
		return c->leaveContext();
	}

	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->find(QString(szName.ptr()));

	if(!def)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified toolbar doesn't exist"));
		return c->leaveContext();
	}

	KviScriptToolBar * pBar;
	
	if(c->hasSwitch('c'))
	{
		pBar = def->instance(c->window()->frame());
	} else {
		for(KviFrame * f = g_pApp->frameList()->first();f;f = g_pApp->frameList()->next())
		{
			pBar = def->instance(f);
		}
	}
#endif // !COMPILE_SCRIPTOOLBAR
	return c->leaveContext();
}


/*
	@doc: toolbar.hide
	@type:
		command
	@title:
		toolbar.hide
	@short:
		Hides a toolbar
	@syntax:
		toolbar.hide [-q] [-a] <name>
	@description:
		Hides the toolbar named <name>.[br]
		If -a is specified then the toolbar is hidden in all the frames, otheriwse it is
		hidden in the current frame only.[br]
		If -q is specified the command runs in quiet mode and does not complain if the
		specified toolbar doesn't exist.[br]
		Please note that this command doesn't destroy the toolbar definition: the toolbar
		can be shown again by using [cmd]toolbar.show[/cmd].[br]
		If you want to completly remove the toolbar definition use [cmd]toolbar.destroy[/cmd].[br]
		If the feature "scripttoolbar" wasn't enabled at compile time then this command does nothing.
	@seealso:
		[fnc]$toolbar.exists[/fnc]()
		[cmd]toolbar.create[/cmd]
		[cmd]toolbar.destroy[/cmd]
*/


static bool toolbar_module_cmd_hide(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::hide");
	KviStr szName;
	if(!g_pUserParser->parseCmdFinalPart(c,szName))return false;

#ifdef COMPILE_SCRIPTTOOLBAR
	if(szName.isEmpty())
	{
		if(!c->hasSwitch('q'))c->warning(__tr("No toolbar name specified"));
		return c->leaveContext();
	}

	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->find(QString(szName.ptr()));

	if(!def)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified toolbar doesn't exist"));
		return c->leaveContext();
	}

	KviScriptToolBar * pBar;
	
	if(c->hasSwitch('c'))
	{
		if(pBar = def->existingInstance(c->window()->frame()))delete pBar;
	} else {
		for(KviFrame * f = g_pApp->frameList()->first();f;f = g_pApp->frameList()->next())
		{
			if(pBar = def->existingInstance(f))delete pBar;
		}
	}
#endif // !COMPILE_SCRIPTOOLBAR
	return c->leaveContext();
}



/*
	@doc: toolbar.exists
	@type:
		function
	@title:
		$toolbar.exists
	@short:
		Checks if a toolbar exists
	@syntax:
		$toolbar.exists(<name>)
	@description:
		Returns 1 if the toolbar named <name> has been defined.[br]
		Note that this doesn't mean that the toolbar is visible!
		If the feature "scripttoolbar" wasn't enabled at compile time then this function returns always 0
	@seealso:
		[cmd]toolbar.create[/cmd]
*/

static bool toolbar_module_fnc_exists(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	KviStr szName = parms->safeFirstParam();
	buffer.append(g_pScriptToolBarManager->find(QString(szName.ptr())) ? '1' : '0');
#else // !COMPILE_SCRIPTTOOLBAR
	buffer.append('0');
#endif // !COMPILE_SCRIPTTOOLBAR
	return true;
}


/*
	@doc: toolbar.list
	@type:
		function
	@title:
		$toolbar.list
	@short:
		Returns a list of defined toolbars
	@syntax:
		$toolbar.list()
	@description:
		Returns a list of defined toolbar names.[br]
		You can use this function to loop thru all the toolbar definitions.[br]
	@seealso:
*/

static bool toolbar_module_fnc_list(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	QDictIterator<KviScriptToolBarDefinition> it(*(g_pScriptToolBarManager->toolBarDefinitions()));
	bool bFirst = true;
	while(KviScriptToolBarDefinition * d = it.current())
	{
		if(bFirst)
		{
			buffer.append(d->name());
			bFirst = false;
		} else {
			buffer.append(',');
			buffer.append(d->name());
		}
		++it;
	}
#endif // !COMPILE_SCRIPTTOOLBAR
	return true;
}



/*
	@doc: toolbar.addbutton
	@type:
		command
	@title:
		toolbar.addbutton
	@keyterms:
		Creating toolbar buttons
	@short:
		Creates a toolbar button
	@syntax:
		toolbar.addbutton [-q] [-d] (<toolbar>,<name>[,<icon>,<text>])<callback command>
	@description:
		Creates a toolbar button named <name> and adds it to the toolbar named <toolbar>.[br]
		The toolbar must be existing but not necessairly visible.[br]
		The <icon> parameter must be a valid [doc:image_id]image identifier[/doc] and defines
		the image shown in the button.[br]
		<text> will be used as "tooltip" text for the button and should contain a short
		description of the button functionality.[br]
		[callback command] is the command that will be called on button press (bound to the
		active window).[br]
		The [-d] switch causes the button to be disabled.[br]
		If the button with the same <name> already exists in the <toolbar> then its
		icon, tooltip text and callback command are updated: this means that you can safely
		use this command to update some of the button features. [cmd]toolbar.setbutton[/cmd]
		is faster function for updating the icon the text or enable state (eventually in a single frame),
		but the only way to change the button code is to use this command.[br]
		The button can be removed from the toolbar by using [cmd]toolbar.removeitem[/cmd]
		-q causes the command to run in quiet mode.[br]
	@seealso:
		[cmd]toolbar.setbutton[/cmd]
		[cmd]toolbar.removeitem[/cmd]
*/

static bool toolbar_module_cmd_addbutton(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::addbutton");

	KviParameterList paramList;
	paramList.setAutoDelete(true);

	KviStr cmd;

	if(!g_pUserParser->parseCallbackCommand(c,&paramList,&cmd))return false;


#ifdef COMPILE_SCRIPTTOOLBAR

	KviStr toolbar = paramList.safeFirstParam();
	KviStr name    = paramList.safeNextParam();
	KviStr icon    = paramList.safeNextParam();
	KviStr label   = paramList.safeNextParam();

	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->find(QString(toolbar.ptr()));

	if(!def)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified toolbar doesn't exist (%s)"),toolbar.ptr());
		return c->leaveContext();
	}

	KviScriptToolBarItemDefinition * i = def->getItemDefinition(QString(name.ptr()),KviScriptToolBarItemDefinition::Button);

	KviCommandFormatter::bufferFromBlock(cmd);

	if(i->type() != KviScriptToolBarItemDefinition::Button)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified name is already assigned to a different type item (%s)"),name.ptr());
		return c->leaveContext();
	}

	if(icon.hasData())i->setIcon(QString(icon.ptr()));
	if(label.hasData())i->setText(QString(label.ptr()));
	i->setCode(QString(cmd.ptr()));
	i->setEnabled(!c->hasSwitch('d'));
#endif // COMPILE_SCRIPTOOLBAR

	return c->leaveContext();
}

/*
	@doc: toolbar.removeitem
	@type:
		command
	@title:
		toolbar.removeitem
	@keyterms:
		Removing toolbar buttons or separators
	@short:
		Removes a toolbar button or separator
	@syntax:
		toolbar.removeitem [-q] <toolbar> <name>
	@description:
		Removes the button named <name> from the toolbar named <toolbar>.[br]
		If no such button exists in the specified toolbar this command does nothing.[br]
		If the <toolbar> parameter doesn't identify an existing toolbar then
		a warning is printed unless the -q switch is used.[br]
	@seealso:
		[cmd]toolbar.addbutton[/cmd]
		[cmd]toolbar.addseparator[/cmd]
*/

static bool toolbar_module_cmd_removeitem(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::removeitem");
	KviStr szName;
	KviStr szToolbar;
	if(!g_pUserParser->parseCmdSingleToken(c,szToolbar))return false;
	if(!g_pUserParser->parseCmdFinalPart(c,szName))return false;
#ifdef COMPILE_SCRIPTTOOLBAR
	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->find(QString(szToolbar.ptr()));

	if(!def)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified toolbar doesn't exist (%s)"),szToolbar.ptr());
		return c->leaveContext();
	}

	def->removeItemDefinition(QString(szName.ptr()));
#endif // COMPILE_SCRIPTTOOLBAR

	return c->leaveContext();
}


/*
	@doc: toolbar.setbutton
	@type:
		command
	@title:
		toolbar.setbutton
	@short:
		Sets the text, the icon or the state of a toolbar button
	@syntax:
		toolbar.setbutton [-a] [-c] [-t | -i] [-e] [-d] [-q] <toolbar> <name> [param]
	@description:
		Changes the state of the button named <name> residing in the toolbar <toolbar>.[br]
		If the specified button or toolbar do not exist then a warning is printed unless
		the -q switch is used.[br]
		-e causes the button to be enabled.[br]
		-d causes the button to be disabled.[br]
		-t sets the button text to [param].[br]
		-i sets the button icon to [param] that must be a valid [doc:image_id]image identifier[/doc].[br]
		-t and -i are mutually exclusive: if both are given then -t takes precedence.[br]
		-e and -d work for the button in the current frame unless -a is specified: in that
		case the button is updated in all the frames.[br]
		-e and -d are mutually exclusive: if both are given then -d takes precedence.[br]
	@seealso:
		[cmd]toolbar.addbutton[/cmd]
*/

static bool toolbar_module_cmd_setbutton(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::setbutton");
	KviStr szName;
	KviStr szToolbar;
	KviStr szText;
	if(!g_pUserParser->parseCmdSingleToken(c,szToolbar))return false;
	if(!g_pUserParser->parseCmdSingleToken(c,szName))return false;
	if(!g_pUserParser->parseCmdFinalPart(c,szText))return false;
#ifdef COMPILE_SCRIPTTOOLBAR
	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->find(QString(szToolbar.ptr()));

	if(!def)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified toolbar doesn't exist (%s)"),szToolbar.ptr());
		return c->leaveContext();
	}

	KviScriptToolBarItemDefinition * i = def->findItemDefinition(QString(szName.ptr()));

	if(!i)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified button doesn't exist (%s)"),szName.ptr());
		return c->leaveContext();
	}
	
	if(i->type() != KviScriptToolBarItemDefinition::Button)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified name is assigned to an item that is not a button (%s)"),szName.ptr());
		return c->leaveContext();
	}
	
	if(c->hasSwitch('t'))i->setText(QString(szText.ptr()));
	else if(c->hasSwitch('i'))i->setIcon(QString(szText.ptr()));
	
	if(c->hasSwitch('e'))i->setEnabled(true,c->hasSwitch('a') ? 0 : c->window()->frame());
	if(c->hasSwitch('d'))i->setEnabled(false,c->hasSwitch('a') ? 0 : c->window()->frame());
	
#endif // COMPILE_SCRIPTTOOLBAR

	return c->leaveContext();
}



/*
	@doc: toolbar.addseparator
	@type:
		command
	@title:
		toolbar.addseparator
	@short:
		Adds a separator to a toolbar
	@syntax:
		toolbar.addseparator [-q] <toolbar> <name>
	@description:
		Adds a separator to the toolbar identified by <toolbar>.[br]
		The separator will have an associated <name> that can be used later to remove it
		by using [cmd]toolbar.removeitem[/cmd].[br]
		-q causes the command to run in silent mode.[br]
	@seealso:
		[cmd]toolbar.addbutton[/cmd]
		[cmd]toolbar.removeitem[/cmd]
*/

static bool toolbar_module_cmd_addseparator(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbar::addseparator");
	KviStr szName;
	KviStr szToolbar;

	if(!g_pUserParser->parseCmdSingleToken(c,szToolbar))return false;
	if(!g_pUserParser->parseCmdFinalPart(c,szName))return false;
#ifdef COMPILE_SCRIPTTOOLBAR
	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->find(QString(szToolbar.ptr()));

	if(!def)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified toolbar doesn't exist (%s)"),szToolbar.ptr());
		return c->leaveContext();
	}

	KviScriptToolBarItemDefinition * i = def->getItemDefinition(QString(szName.ptr()),KviScriptToolBarItemDefinition::Separator);


	if(i->type() != KviScriptToolBarItemDefinition::Separator)
	{
		if(!c->hasSwitch('q'))c->warning(__tr("The specified name is assigned to an item that is not a separator (%s)"),szName.ptr());
		return c->leaveContext();
	}
	
#endif // COMPILE_SCRIPTTOOLBAR

	return c->leaveContext();
}




static bool toolbar_module_init(KviModule * m)
{
	// setlabel , $label , $position , move , $itempos , $itemexists , $itemtype

	m->registerCommand("create",toolbar_module_cmd_create);
	m->registerCommand("show",toolbar_module_cmd_show);
	m->registerCommand("hide",toolbar_module_cmd_hide);
	m->registerCommand("destroy",toolbar_module_cmd_destroy);
	m->registerCommand("addbutton",toolbar_module_cmd_addbutton);
	m->registerCommand("addseparator",toolbar_module_cmd_addseparator);
	m->registerCommand("removeitem",toolbar_module_cmd_removeitem);
	m->registerCommand("setbutton",toolbar_module_cmd_setbutton);
	m->registerCommand("define",toolbar_module_cmd_define);

	m->registerFunction("exists",toolbar_module_fnc_exists);
	m->registerFunction("list",toolbar_module_fnc_list);

	return true;
}

static bool toolbar_module_cleanup(KviModule *m)
{
	return true;
}

KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"toolbar",                                                 // module name
	"1.0.0",                                                // module version
	"Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)", // author & (C)
	"Interface to the scriptable toolbars",
	toolbar_module_init,
	0,
	0,
	toolbar_module_cleanup
};
