//=================================================================================
//
//   File : kvi_window.cpp
//   Creation date : Sat Sep 01 2001 17:13:12 CEST by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2001 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=================================================================================

#include "kvi_module.h"
#include "kvi_uparser.h"
#include "kvi_console.h"
#include "kvi_options.h"
#include "kvi_irclink.h"
#include "kvi_frame.h"
#include "kvi_locale.h"
#include "kvi_app.h"
#include "kvi_error.h"
#include "kvi_ircview.h"
#include "kvi_input.h"

#include <qasciidict.h>

// kvi_app.cpp
extern KVIRC_API QAsciiDict<KviWindow> * g_pGlobalWindowDict;

// $window.caption $window.x $window.y $window.width $window.height $window.isActive $window.type
// $window.input.text $window.input.cursorpos $window.input.textlen

static KviWindow * get_window_by_first_param(KviCommand *c,KviParameterList *parms)
{
	KviStr * id = parms->safeFirst();
	if(id->isEmpty())return c->window();
	return g_pApp->findWindow(id->ptr());
}

#define GET_WINDOW_BY_FINAL_PART \
	KviStr windowid; \
	if(!g_pUserParser->parseCmdFinalPart(c,windowid))return false; \
	KviWindow * wnd = 0; \
	if(windowid.isEmpty()) \
		wnd = c->window(); \
	else \
	{ \
		wnd = g_pApp->findWindow(windowid.ptr()); \
		if(!wnd && !c->hasSwitch('q'))c->warning(__tr("The window with id '%s' does not exist"),windowid.ptr()); \
	}

/*
	@doc: window.clearOutput
	@type:
		command
	@title:
		window.clearOutput
	@short:
		Clears the output a window
	@syntax:
		window.clearOutput [-q] [window_id]
	@description:
		Clears the text output of the window specified by window_id. If window_id is missing then
		the current window is used. If the window has no text output then no operations is performed.
		If the specified window does not exist a warning is printed unless the -q switch is used.
	@seealso:
		[fnc]$window.hasOutput[/fnc]()
*/

static bool window_module_cmd_clearOutput(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::clearOutput");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)
	{
		if(wnd->view())wnd->view()->emptyBuffer(true);
	}
	return c->leaveContext();
}

/*
	@doc: window.close
	@type:
		command
	@title:
		window.close
	@short:
		Closes a window
	@syntax:
		window.close [-q] [window_id]
	@description:
		Closes the window specified by window_id. If window_id is missing then
		the current window is closed. The close operation is asynchronous: it is
		performed immediately after the script has terminated the execution and
		the control is returned to the main KVIrc core. If the specified window
		does not exist a warning is printed unless the -q switch is used.
*/

static bool window_module_cmd_close(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::close");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->delayedClose();
	return c->leaveContext();
}

/*
	@doc: window.dock
	@type:
		command
	@title:
		window.dock
	@short:
		Docks a window
	@syntax:
		window.dock [-q] [window_id]
	@description:
		Docks the window specified by window_id. If window_id is missing then
		the current window is docked. If the specified window was already docked then
		no operation is performed. If the specified window
		does not exist a warning is printed unless the -q switch is used.
	@seealso:
		[cmd]window.undock[/cmd]
*/

static bool window_module_cmd_dock(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::dock");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->dock();
	return c->leaveContext();
}

/*
	@doc: window.undock
	@type:
		command
	@title:
		window.undock
	@short:
		Undocks a window
	@syntax:
		window.undock [-q] [window_id]
	@description:
		Undocks the window specified by window_id. If window_id is missing then
		the current window is undocked. If the specified window was already undocked then
		no operation is performed. If the specified window
		does not exist a warning is printed unless the -q switch is used.
	@seealso:
		[cmd]window.dock[/cmd]
*/

static bool window_module_cmd_undock(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::undock");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->undock();
	return c->leaveContext();
}


/*
	@doc: window.maximize
	@type:
		command
	@title:
		window.maximize
	@short:
		Maximizes a window
	@syntax:
		window.maximize [-q] [window_id]
	@description:
		Maximizes the window specified by window_id. If window_id is missing then
		the current window is maximized. If the specified window was already maximized then
		no operation is performed. If the specified window
		does not exist a warning is printed unless the -q switch is used.
	@seealso:
		[cmd]window.minimize[/cmd], [cmd]window.restore[/cmd], [fnc]$window.isMaxmimized[/cmd],
		[fnc]$window.isMinimized[/cmd]
*/

static bool window_module_cmd_maximize(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::maximize");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->maximize();
	return c->leaveContext();
}

/*
	@doc: window.minimize
	@type:
		command
	@title:
		window.minimize
	@short:
		Minimizes a window
	@syntax:
		window.minimize [-q] [window_id]
	@description:
		Minimizes the window specified by window_id. If window_id is missing then
		the current window is minimized. If the specified window was already minimized then
		no operation is performed. If the specified window
		does not exist a warning is printed unless the -q switch is used.
	@seealso:
		[cmd]window.maximize[/cmd], [cmd]window.restore[/cmd], [fnc]$window.isMaxmimized[/cmd],
		[fnc]$window.isMinimized[/cmd]
*/

static bool window_module_cmd_minimize(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::minimize");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->minimize();
	return c->leaveContext();
}

/*
	@doc: window.restore
	@type:
		command
	@title:
		window.restore
	@short:
		Restores a window
	@syntax:
		window.restore [-q] [window_id]
	@description:
		Restores the window specified by window_id. If window_id is missing then
		the current window is restored. If the specified window was already restored then
		no operation is performed. If the specified window
		does not exist a warning is printed unless the -q switch is used.
	@seealso:
		[cmd]window.maximize[/cmd], [cmd]window.minimize[/cmd], [fnc]$window.isMaxmimized[/cmd],
		[fnc]$window.isMinimized[/cmd]
*/

static bool window_module_cmd_restore(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::restore");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->restore();
	return c->leaveContext();
}

/*
	@doc: window.activate
	@type:
		command
	@title:
		window.activate
	@short:
		Activates a window
	@syntax:
		window.activate [-q] [window_id]
	@description:
		Activates the window specified by window_id. If window_id is missing then
		the current window is activated. If the specified window
		does not exist a warning is printed unless the -q switch is used.
		Please note that if the window is currently docked to a frame then this
		command will NOT raise the frame window. If you're interested in
		the user's attention then you might be interested in [cmd]window.demandAttention[/cmd].
	@seealso:
		[cmd]window.demandAttention[/cmd]
*/

static bool window_module_cmd_activate(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::activate");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->autoRaise();
	return c->leaveContext();
}

/*
	@doc: window.demandAttention
	@type:
		command
	@title:
		window.demandAttention
	@short:
		Flashes a window's system taskbar entry
	@syntax:
		window.demandAttention [-q] [window_id]
	@description:
		Flashes the system taskbar entry of the window
		specified by the window_id. If window_id is missing then
		the current window's system taskbar entry is flashed. If the specified window
		does not exist a warning is printed unless the -q switch is used.
		If the window is currently docked in a frame then the frame's
		system taskbar entry will be flashed.
		Please note that this command is highly system dependant:
		on systems that do not have a system taskbar or there
		is no way to flash an etry this command will do nothing.
		At the time of writing this command works flawlessly on
		Windows and in KDE compilations.
	@seealso:
		[cmd]window.demandAttention[/cmd]
*/

static bool window_module_cmd_demandAttention(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window::demandAttention");
	GET_WINDOW_BY_FINAL_PART
	if(wnd)wnd->demandAttention();
	return c->leaveContext();
}

/*
	@doc: window.isMaximized
	@type:
		function
	@title:
		$window.isMaximized
	@short:
		Checks if a window is currently maximized
	@syntax:
		$window.isMaximized
		$window.isMaximized(<window_id>)
	@descritpion:
		Returns 1 if the window specified by <window_id> is currently maximized and 0 otherwise.
		The form with no parameters works on the current window. If the specified window
		doesn't exist then 0 is returned.
	@seealso:
		[fnc]$window.isMinimized[/fnc](), [cmd]window.maximize[/cmd],
		[cmd]window.minimize[/cmd], [cmd]window.restore[/cmd]
*/


static bool window_module_fnc_isMaximized(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"$window.isMaximized");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)buffer.append(wnd->isMaximized() ? '1' : '0');
	else buffer.append('0');
	return c->leaveContext();
}


/*
	@doc: window.isMinimized
	@type:
		function
	@title:
		$window.isMinimized
	@short:
		Checks if a window is currently minimized
	@syntax:
		$window.isMinimized
		$window.isMinimized(<window_id>)
	@descritpion:
		Returns 1 if the window specified by <window_id> is currently minimized and 0 otherwise.
		The form with no parameters works on the current window. If the specified window
		doesn't exist then 0 is returned.
	@seealso:
		[fnc]$window.isMaximized[/fnc](), [cmd]window.maximize[/cmd],
		[cmd]window.minimize[/cmd], [cmd]window.restore[/cmd]
*/


static bool window_module_fnc_isMinimized(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"$window.isMinimized");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)buffer.append(wnd->isMinimized() ? '1' : '0');
	else buffer.append('0');
	return c->leaveContext();
}


/*
	@doc: window.hasInput
	@type:
		function
	@title:
		$window.hasInput
	@short:
		Checks if a window has an input field
	@syntax:
		$window.hasInput
		$window.hasInput(<window_id>)
	@descritpion:
		Returns 1 if the window specified by <window_id> has an input field and 0 otherwise.
		The form with no parameters works on the current window. If the specified window
		doesn't exist then 0 is returned.
	@seealso:
		[fnc]$window.hasOutput[/fnc]
*/


static bool window_module_fnc_hasInput(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"$window.hasInput");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)buffer.append(wnd->input() ? '1' : '0');
	else buffer.append('0');
	return c->leaveContext();
}


/*
	@doc: window.hasUserFocus
	@type:
		function
	@title:
		$window.hasUserFocus
	@short:
		Checks if a window has the user focus
	@syntax:
		$window.hasUserFocus
		$window.hasUserFocus(<window_id>)
	@descritpion:
		Returns 1 if the window specified by <window_id> has
		currently the user focus and 0 otherwise.
		The form with no parameters works on the current window.
		If the specified window doesn't exist then 0 is returned.
		A window has the user focus if it is the KVIrc's active
		window and has the user's input focus (i.e. typing
		on the keyboard will write in this window).
	@seealso:
*/

static bool window_module_fnc_hasUserFocus(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"$window.hasUserFocus");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)
	{
		bool b = (wnd == g_pApp->activeWindow()) && wnd->isActiveWindow();
		buffer.append(b ? '1' : '0');
	} else buffer.append('0');
	return c->leaveContext();
}

/*
	@doc: window.console
	@type:
		function
	@title:
		$window.console
	@short:
		Returns the console that a window is attacched to
	@syntax:
		$window.console
		$window.console(<window_id>)
	@descritpion:
		Returns the id of the console window that the window specified by window_id is attacched to.
		The console is the main (and only) console of the IRC context. If window_id is missing then
		the current window is used. If this window does not belong to an irc context (and thus has
		no attacched console) then 0 is returned.
	@seealso:
*/


static bool window_module_fnc_console(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"window::$console");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)buffer.append(wnd->console() ? wnd->console()->id() : "0");
	else buffer.append("0");
	return c->leaveContext();
}

/*
	@doc: window.hasOutput
	@type:
		function
	@title:
		$window.hasOutput
	@short:
		Checks if a window has a text output widget
	@syntax:
		$window.hasOutput
		$window.hasOutput(<window_id>)
	@descritpion:
		Returns 1 if the window specified by <window_id> has a text output widget and 0 otherwise.
		The form with no parameters works on the current window. If the specified window
		doesn't exist then 0 is returned.
	@seealso:
		[fnc]$window.hasInput[/fnc]
*/


static bool window_module_fnc_hasOutput(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"window::$hasOutput");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)buffer.append(wnd->view() ? '1' : '0');
	else buffer.append('0');
	return c->leaveContext();
}

/*
	@doc: window.exists
	@type:
		function
	@title:
		$window.exists
	@short:
		Checks for the existence of a window
	@syntax:
		$window.exists(<window_id>)
	@description:
		Returns 1 if a specified window exists
	@seealso:
*/

static bool window_module_fnc_exists(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	buffer.append(g_pApp->findWindow(parms->safeFirstParam()) ? '1' : '0');
	return true;
}


/*
	@doc: window.type
	@type:
		function
	@title:
		$window.type
	@short:
		Returns the type of a window
	@syntax:
		$window.type
		$window.type(<window_id>)
	@descritpion:
		Returns the type of the window with <window_id>.[br]
		The form with no parameters returns the type of the current window.[br]
		If the window with the specified id does not exist, an empty string is returned.[br]
	@seealso:
*/


static bool window_module_fnc_type(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"$window.type");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)buffer.append(wnd->typeString());
	return c->leaveContext();
}



/*
	@doc: window.caption
	@type:
		function
	@title:
		$window.caption
	@short:
		Returns the caption of a window
	@syntax:
		$window.caption
		$window.caption(<window_id>)
	@descritpion:
		Returns the caption of the window with <window_id>.[br]
		The form with no parameters returns the caption of the current window.[br]
		If the window with the specified id does not exist, an empty string is returned.[br]
	@seealso:
*/


static bool window_module_fnc_caption(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"$window.caption");
	KviWindow * wnd = get_window_by_first_param(c,parms);
	if(wnd)buffer.append(wnd->plainTextCaption());
	return c->leaveContext();
}

/*
	@doc: window.listtypes
	@type:
		command
	@title:
		window.listtypes
	@short:
		Lists available types of windows
	@syntax:
		window.listtypes
	@description:
		Lists the types of windows that are built in the current release of KVIrc.[br]
		This is actually a command and not a static list just because new window
		types may be added in subsequent releases.[br]
	@seealso:
		[cmd]window.listtypes[/cmd]
*/

static bool window_module_cmd_listtypes(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"window.listtypes");

	KviStr dummy;
	if(!g_pUserParser->parseCmdFinalPart(c,dummy))return false;
	c->window()->listWindowTypes();
	return c->leaveContext();
}


/*
	@doc: window.list
	@type:
		function
	@title:
		$window.list
	@short:
		Generates lists of windows
	@syntax:
		$window.list(<type>[,<irc_context_id>])
	@description:
		Returns an array of window identifiers with a speficied type and eventually belonging to a specified
		irc context.[br]
		<type> is a window type such as 'query' or 'channel'.[br]
		See [cmd]window.listtypes[/cmd] for a list of available window types in this KVIrc release.[br]
		If <type> is the special word 'all', all the window types are listed.[br]
		<irc_context_id> specifies the irc context in which the windows are searched.[br]
		If no <irc_context_id> is specified , the current one is used.[br]
		If <irc_context_id> is the special word 'all' , all the irc context are searched.[br]
		The windows that do not belong to any irc context (such as DCC windows), must be searched
		by using 'all' as <irc_context_id>.
	@examples:
		[example]
			[comment]# List all the queries of the current irc context[/comment]
			[cmd]echo[/cmd] $window.list(query)
			[comment]# Equivalent to the above[/comment]
			[cmd]echo[/cmd] $window.list(query,[fnc]$ic[/fnc])
			[comment]# List all the channels in all the irc contexts[/comment]
			[cmd]echo[/cmd] $window.list(channel,all)
			[comment]# List all the windows in the current irc context[/comment]
			[cmd]echo[/cmd] $window.list(all)
			[comment]# List all the windows in all irc contexts[/comment]
			[cmd]echo[/cmd] $window.list(all,all)
			[comment]# List all the DCC Send windows[/comment]
			[cmd]echo[/cmd] $window.list(dccsend,all)
			[comment]# Ok , let's use it[/comment]
			[comment]# A nice alias that allows iterating commands through all the consoles[/comment]
			[comment]# Note the array returned by $window.list[/comment]
			[comment]# This is by LatinSuD :)[/comment]
			[cmd]alias[/cmd](iterate)
			{
				%ctxt[]=[fnc]$window.list[/fnc](console,all)
				[cmd]for[/cmd](%i=0;%i<%ctxt[]#;%i++)
				{
					[cmd]eval[/cmd] -r=%ctxt[%i] $0-
				}
			}
			iterate [cmd]echo[/cmd] Hi ppl! :)
			[comment]# The returned array works nicely also in [cmd]foreach[/cmd][/comment]
			[comment]# Say hi to all the channels :)[/comment]
			[cmd]alias[/cmd](sayallchans)
			{
				[cmd]foreach[/cmd](%x,[fnc]$window.list[/fnc](channel,all))
						[cmd]say[/cmd] -r=%x $0-;
			}
			sayallchans Hi ppl :)
		[/example]
	@seealso:
		[cmd]window.listtypes[/cmd]
*/

static bool window_module_fnc_list(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"$window.list");

	// all , all
	// all , context
	// type , all
	// type , context

	KviStr * type = parms->safeFirst();
	KviStr * cntx = parms->safeNext();

	if(type->isEmpty())return c->error(KviError_notEnoughParameters,__tr("Window type or 'all' expected as first parameter"));

	c->beginListArrayOrDictionaryReturnIdentifier();

	int idx = 0;

	if(kvi_strEqualCI(cntx->ptr(),"all"))
	{
		// all contexts
		bool bAllWindows = kvi_strEqualCI(type->ptr(),"all");
		QAsciiDictIterator<KviWindow> it(*g_pGlobalWindowDict);

		while(KviWindow * wnd = it.current())
		{
			if(bAllWindows)
			{
				c->addListArrayOrDictionaryReturnValue(idx++,wnd->id(),buffer);
			} else {
				if(kvi_strEqualCI(wnd->typeString(),type->ptr()))
				{
					c->addListArrayOrDictionaryReturnValue(idx++,wnd->id(),buffer);
				}
			}
			++it;
		}
	} else {
		// some specified context
		unsigned int uId = 0;

		if(cntx->hasData())
		{
			// specific context
			bool bOk;
			uId = cntx->toUInt(&bOk);
			if(!bOk)return c->error(KviError_invalidParameter,__tr("Invalid IRC context id '%s'"),cntx->ptr());
		} else {
			// current irc context
			if(!c->window()->console())return c->noIrcContext();
			uId = c->window()->console()->ircContextId();
		}

		bool bAllWindows = kvi_strEqualCI(type->ptr(),"all");
		QAsciiDictIterator<KviWindow> it(*g_pGlobalWindowDict);

		while(KviWindow * wnd = it.current())
		{
			if(wnd->console())
			{
				if(wnd->console()->ircContextId() == uId)
				{
					if(bAllWindows)
					{
						c->addListArrayOrDictionaryReturnValue(idx++,wnd->id(),buffer);
					} else {
						if(kvi_strEqualCI(wnd->typeString(),type->ptr()))
						{
							c->addListArrayOrDictionaryReturnValue(idx++,wnd->id(),buffer);
						}
					}
				}
			}
			++it;
		}
	}

	return c->leaveContext();
}



static bool window_module_init(KviModule *m)
{
	m->registerFunction("list",window_module_fnc_list);
	m->registerCommand("listtypes",window_module_cmd_listtypes);

	m->registerFunction("exists",window_module_fnc_exists);
	m->registerFunction("type",window_module_fnc_type);
	m->registerFunction("caption",window_module_fnc_caption);
	m->registerFunction("isMaximized",window_module_fnc_isMaximized);
	m->registerFunction("isMinimized",window_module_fnc_isMinimized);
	m->registerFunction("hasInput",window_module_fnc_hasInput);
	m->registerFunction("hasOutput",window_module_fnc_hasOutput);
	m->registerFunction("hasUserFocus",window_module_fnc_hasUserFocus);
	m->registerFunction("console",window_module_fnc_console);
	
	m->registerCommand("close",window_module_cmd_close);
	m->registerCommand("activate",window_module_cmd_activate);
	m->registerCommand("dock",window_module_cmd_dock);
	m->registerCommand("undock",window_module_cmd_undock);
	m->registerCommand("maximize",window_module_cmd_maximize);
	m->registerCommand("minimize",window_module_cmd_minimize);
	m->registerCommand("restore",window_module_cmd_restore);
	m->registerCommand("clearOutput",window_module_cmd_clearOutput);
	m->registerCommand("demandAttention",window_module_cmd_demandAttention);

	// saveOutput (view()->saveBuffer())
/*
	m->registerFunction("geometry",window_module_fnc_geometry);
	m->registerCommand("setGeometry",window_module_cmd_setGeometry);

	m->registerFunction("isActive",window_module_fnc_isActive);

	m->registerCommand("setCaption",window_module_cmd_setCaption);

	// Input stuff
	m->registerFunction("inputText",window_module_fnc_inputText);
	m->registerCommand("setInputText",window_module_cmd_setInputText);

	// Output stuff
	m->registerFunction("outputBufferSize",window_module_fnc_outputBufferSize);
	m->registerCommand("setOutputBufferSize",window_module_cmd_setOutputBufferSize);

	m->registerFunction("outputShowsImages",window_module_fnc_outputShowsImages);
	m->registerCommand("setOutputShowsImages",window_module_cmd_setOutputShowsImages);
	
	m->registerFunction("outputTimestamp",window_module_fnc_outputTimestamp);
	m->registerCommand("setOutputTimestamp",window_module_cmd_setOutputTimestamp);

	m->registerFunction("hasOutputBackgroundPixmap",window_module_fnc_hasOutputBackgroundPixmap);
	m->registerCommand("setOutputBackgroundPixmap",window_module_cmd_setOutputBackgroundPixmap);

*/
	return true;
}

static bool window_module_cleanup(KviModule *m)
{
	return true;
}

static bool window_module_can_unload(KviModule *m)
{
	return true;
}

KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"Window",                                               // module name
	"1.0.0",                                                // module version
	"Copyright (C) 2001-2004 Szymon Stefanek (pragma at kvirc dot net)", // author & (C)
	"KVIrc window management functions",
	window_module_init,
	window_module_can_unload,
	0,
	window_module_cleanup
};
